################################################################################
# Copyright (c) 2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpmodels.fetch.aiohttp`."""

import hashlib
import types
from typing import List

import h5py
import pytest
import aiohttp
import yarl
import katsdptelstate.aio.memory

from katsdpmodels import models, fetch
import katsdpmodels.fetch.aiohttp as fetch_aiohttp
from test_utils import get_data, get_data_url, get_file_url, DummyModel


pytestmark = [pytest.mark.asyncio]


@pytest.mark.parametrize('use_file', [True, False])
@pytest.mark.parametrize('filename', ['rfi_mask_ranges.h5', 'direct.alias', 'indirect.alias'])
async def test_fetch_model_simple(use_file, filename, web_server) -> None:
    url = get_file_url(filename) if use_file else web_server(filename)
    with await fetch_aiohttp.fetch_model(url, DummyModel) as model:
        assert len(model.ranges) == 2
        assert not model.is_closed
    assert model.is_closed


async def test_fetch_model_retry(web_server) -> None:
    """Test that retry happens on a 5xx server error."""
    url = web_server('rfi_mask_ranges.h5').replace('/static/', '/failonce/')
    with await fetch_aiohttp.fetch_model(url, DummyModel) as model:
        assert len(model.ranges) == 2
        assert not model.is_closed
    assert model.is_closed


async def test_fetch_model_alias_loop(web_server) -> None:
    url = web_server('loop.alias')
    with pytest.raises(models.TooManyAliasesError) as exc_info:
        await fetch_aiohttp.fetch_model(url, DummyModel)
    assert exc_info.value.url == url
    assert exc_info.value.original_url == url


async def test_fetch_model_too_many_aliases(web_server, monkeypatch) -> None:
    monkeypatch.setattr(fetch, 'MAX_ALIASES', 1)
    with pytest.raises(models.TooManyAliasesError):
        await fetch_aiohttp.fetch_model(web_server('indirect.alias'), DummyModel)
    # Check that 1 level of alias is still permitted
    with await fetch_aiohttp.fetch_model(web_server('direct.alias'), DummyModel):
        pass


async def test_fetch_model_absolute_alias(web_server) -> None:
    url = web_server('to_file.alias')
    with pytest.raises(models.AbsoluteAliasError) as exc_info:
        await fetch_aiohttp.fetch_model(url, DummyModel)
    assert exc_info.value.url == url
    assert exc_info.value.original_url == url


@pytest.mark.parametrize('filename', ['bad_model_type.h5', 'no_model_type.h5'])
async def test_fetch_model_model_type_error(filename, web_server) -> None:
    url = web_server(filename)
    with pytest.raises(models.ModelTypeError) as exc_info:
        await fetch_aiohttp.fetch_model(url, DummyModel)
    assert exc_info.value.url == url
    assert exc_info.value.original_url == url
    assert 'rfi_mask' in str(exc_info.value)


async def test_fetch_model_bad_created(web_server) -> None:
    url = web_server('bad_created.h5')
    with pytest.raises(models.DataError, match='Invalid creation timestamp') as exc_info:
        await fetch_aiohttp.fetch_model(url, DummyModel)
    assert exc_info.value.url == url
    assert exc_info.value.original_url == url


@pytest.mark.parametrize('filename', ['bad_model_version.h5', 'no_model_version.h5'])
async def test_fetch_model_model_version_error(filename, web_server) -> None:
    url = web_server(filename)
    with pytest.raises(models.ModelVersionError) as exc_info:
        await fetch_aiohttp.fetch_model(url, DummyModel)
    assert exc_info.value.url == url
    assert exc_info.value.original_url == url
    assert 'model_version' in str(exc_info.value)


async def test_fetch_model_cached_model_type_error(web_server) -> None:
    class OtherModel(models.SimpleHDF5Model):
        model_type = 'other'

        @classmethod
        def from_hdf5(cls, hdf5: h5py.File) -> 'OtherModel':
            return cls()

        def to_hdf5(self, hdf5: h5py.File) -> None:
            pass

    url = web_server('rfi_mask_ranges.h5')
    async with fetch_aiohttp.Fetcher() as fetcher:
        await fetcher.get(url, DummyModel)
        with pytest.raises(models.ModelTypeError) as exc_info:
            await fetcher.get(url, OtherModel)
        assert exc_info.value.url == url
        assert exc_info.value.original_url == url


async def test_fetch_model_bad_content_type(mock_aioresponses) -> None:
    data = get_data('rfi_mask_ranges.h5')
    url = get_data_url('bad_content_type.h5')
    mock_aioresponses.get(url, content_type='image/png', body=data)
    with pytest.raises(models.FileTypeError,
                       match='Expected application/x-hdf5, not image/png') as exc_info:
        async with fetch_aiohttp.Fetcher() as fetcher:
            await fetcher.get(url, DummyModel)
    assert exc_info.value.url == url
    assert exc_info.value.original_url == url


async def test_fetch_model_bad_extension(web_server) -> None:
    url = web_server('wrong_extension.blah')
    with pytest.raises(models.FileTypeError) as exc_info:
        async with fetch_aiohttp.Fetcher() as fetcher:
            await fetcher.get(url, DummyModel)
    assert exc_info.value.url == url
    assert exc_info.value.original_url == url


async def test_fetch_model_not_hdf5(web_server) -> None:
    url = web_server('not_hdf5.h5')
    with pytest.raises(models.DataError) as exc_info:
        await fetch_aiohttp.fetch_model(url, DummyModel)
    assert exc_info.value.url == url
    assert exc_info.value.original_url == url


async def test_fetch_model_checksum_ok(mock_aioresponses) -> None:
    data = get_data('rfi_mask_ranges.h5')
    digest = hashlib.sha256(data).hexdigest()
    url = get_data_url(f'sha256_{digest}.h5')
    mock_aioresponses.get(url, content_type='application/x-hdf5', body=data)
    with await fetch_aiohttp.fetch_model(url, DummyModel) as model:
        assert model.checksum == digest


async def test_fetch_model_checksum_bad(mock_aioresponses) -> None:
    data = get_data('rfi_mask_ranges.h5')
    digest = hashlib.sha256(data).hexdigest()
    url = get_data_url(f'sha256_{digest}.h5')
    # Now invalidate the digest
    data += b'blahblahblah'
    mock_aioresponses.get(url, content_type='application/x-hdf5', body=data)
    with pytest.raises(models.ChecksumError) as exc_info:
        await fetch_aiohttp.fetch_model(url, DummyModel)
    assert exc_info.value.url == url


@pytest.mark.parametrize('filename', ['does_not_exist.h5', 'does_not_exist.alias'])
async def test_fetch_model_bad_http_status(filename, web_server) -> None:
    url = web_server(filename)
    with pytest.raises(aiohttp.ClientResponseError) as exc_info:
        await fetch_aiohttp.fetch_model(url, DummyModel)
    assert exc_info.value.status == 404


async def test_fetch_model_http_redirect(mock_aioresponses) -> None:
    url = get_data_url('subdir/redirect.alias')
    # aioresponses doesn't properly handle relative URLs in redirects, so we
    # need to use an absolute URL.
    new_url = get_data_url('direct.alias')
    mock_aioresponses.get(url, headers={'Location': new_url}, status=307)
    with await fetch_aiohttp.fetch_model(url, DummyModel) as model:
        assert len(model.ranges) == 2


async def test_fetch_model_connection_error(mock_aioresponses) -> None:
    # aioresponses raises ConnectionError for any unregistered URL
    with pytest.raises(aiohttp.ClientConnectionError):
        await fetch_aiohttp.fetch_model(get_data_url('does_not_exist.h5'), DummyModel)


async def test_fetcher_caching(mock_aioresponses) -> None:
    async with fetch_aiohttp.Fetcher() as fetcher:
        model1 = await fetcher.get(get_data_url('rfi_mask_ranges.h5'), DummyModel)
        model2 = await fetcher.get(get_data_url('indirect.alias'), DummyModel)
        model3 = await fetcher.get(get_data_url('direct.alias'), DummyModel)
        assert model1 is model2
        assert model1 is model3
        assert not model1.is_closed
    # Not supported by aioresponses
    # assert len(mock_aioresponses.calls) == 3
    assert model1.is_closed


def _tracing_session(urls: List[yarl.URL]) -> aiohttp.ClientSession:
    async def record_urls(session: aiohttp.ClientSession,
                          trace_config_ctx: types.SimpleNamespace,
                          params: aiohttp.TraceRequestStartParams) -> None:
        urls.append(params.url)

    trace_config = aiohttp.TraceConfig()
    trace_config.on_request_start.append(record_urls)     # type: ignore
    return aiohttp.ClientSession(trace_configs=[trace_config])


async def test_fetcher_custom_session(web_server) -> None:
    urls: List[yarl.URL] = []
    async with _tracing_session(urls) as session:
        async with fetch_aiohttp.Fetcher(session=session) as fetcher:
            assert fetcher.session is session
            await fetcher.get(web_server('direct.alias'), DummyModel)
        assert len(urls) == 2
        assert not session.closed
    assert session.closed


async def test_custom_session(web_server) -> None:
    urls: List[yarl.URL] = []
    async with _tracing_session(urls) as session:
        await fetch_aiohttp.fetch_model(web_server('direct.alias'), DummyModel, session=session)
        assert len(urls) == 2
        assert not session.closed
    assert session.closed


async def test_fetcher_resolve(web_server) -> None:
    url = web_server('indirect.alias')
    async with fetch_aiohttp.Fetcher() as fetcher:
        urls = await fetcher.resolve(url)
    assert urls == [
        web_server('indirect.alias'),
        web_server('direct.alias'),
        web_server('rfi_mask_ranges.h5')
    ]


@pytest.fixture
async def telstate(telstate):
    # Extract the memory store from the sync telstate, wrap into an async telstate
    backend = katsdptelstate.aio.memory.MemoryBackend.from_sync(telstate.backend)
    yield katsdptelstate.aio.TelescopeState(backend)
    backend.close()
    await backend.wait_closed()


@pytest.fixture
async def telstate_fetcher(telstate):
    fetcher = fetch_aiohttp.TelescopeStateFetcher(telstate)
    async with fetcher:
        yield fetcher


async def test_telescope_state_fetcher_missing_base_url(telstate) -> None:
    await telstate.delete('sdp_model_base_url')
    async with fetch_aiohttp.TelescopeStateFetcher(telstate) as fetcher:
        with pytest.raises(models.TelescopeStateError, match='not found'):
            await fetcher.get('model_key', DummyModel)


async def test_telescope_state_fetcher_bad_base_url(telstate) -> None:
    await telstate.delete('sdp_model_base_url')
    await telstate.set('sdp_model_base_url', b'Not a string')
    async with fetch_aiohttp.TelescopeStateFetcher(telstate) as fetcher:
        with pytest.raises(models.TelescopeStateError, match='should be a str'):
            await fetcher.get('model_key', DummyModel)


async def test_telescope_state_fetcher_missing_key(telstate_fetcher) -> None:
    with pytest.raises(models.TelescopeStateError, match='not found'):
        await telstate_fetcher.get('missing_key', DummyModel)


async def test_telescope_state_fetcher_bad_key(telstate, telstate_fetcher) -> None:
    await telstate.set('bad_key', 123)
    with pytest.raises(models.TelescopeStateError, match='should be a str'):
        await telstate_fetcher.get('bad_key', DummyModel)


async def test_telescope_state_fetcher_connection_error(telstate_fetcher, mocker) -> None:
    mocker.patch('katsdptelstate.aio.TelescopeState.__getitem__',
                 side_effect=katsdptelstate.ConnectionError('test error'))
    with pytest.raises(models.TelescopeStateError, match='test error'):
        await telstate_fetcher.get('model_key', DummyModel)


async def test_telescope_state_fetcher_good(telstate_fetcher, mock_aioresponses) -> None:
    model = await telstate_fetcher.get('model_key', DummyModel)
    assert len(model.ranges) == 2


async def test_telescope_state_fetcher_override(telstate_fetcher, mock_aioresponses) -> None:
    telstate2 = katsdptelstate.aio.TelescopeState()
    await telstate2.set('another_key', 'rfi_mask_ranges.h5')
    model = await telstate_fetcher.get('another_key', DummyModel, telstate=telstate2)
    assert len(model.ranges) == 2



################################################################################
# Copyright (c) 2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Masks for roll-off at the edges of a band."""

from typing import Type, TypeVar, ClassVar, Any
from typing_extensions import Literal

import numpy as np
import astropy.units as u
import astropy.table
import h5py

from . import models

_B = TypeVar('_B', bound='BandMaskRanges')


class SpectralWindow:
    """Defines a frequency range for use with :class:`BandMask`."""

    def __init__(self, bandwidth: u.Quantity, centre_frequency: u.Quantity):
        self.bandwidth = bandwidth.to(u.Hz)
        self.centre_frequency = centre_frequency.to(u.Hz)

    @property
    def min_frequency(self) -> u.Quantity:
        return self.centre_frequency - self.bandwidth / 2

    @property
    def max_frequency(self) -> u.Quantity:
        return self.centre_frequency + self.bandwidth / 2


class BandMask(models.SimpleHDF5Model):
    model_type: ClassVar[Literal['band_mask']] = 'band_mask'

    def is_masked(self, spectral_window: SpectralWindow, frequency: u.Quantity,
                  channel_width: u.Quantity = 0 * u.Hz) -> Any:
        raise NotImplementedError()      # pragma: nocover

    @classmethod
    def from_hdf5(cls, hdf5: h5py.File) -> 'BandMask':
        model_format = models.get_hdf5_attr(hdf5.attrs, 'model_format', str) or ''
        if model_format == 'ranges':
            return BandMaskRanges.from_hdf5(hdf5)
        else:
            raise models.ModelFormatError(
                f'Unknown model_format {model_format!r} for {cls.model_type}')

    def to_hdf5(self, hdf5: h5py.File) -> None:
        raise NotImplementedError()      # pragma: nocover


class BandMaskRanges(BandMask):
    model_format: ClassVar[Literal['ranges']] = 'ranges'

    def __init__(self, ranges: astropy.table.Table) -> None:
        super().__init__()
        cols = ('min_fraction', 'max_fraction')
        self.ranges = astropy.table.Table(
            [ranges[col] for col in cols],
            names=cols,
            dtype=(np.float64, np.float64)
        )

    def is_masked(self, spectral_window: SpectralWindow, frequency: u.Quantity,
                  channel_width: u.Quantity = 0 * u.Hz) -> Any:
        low_freq = frequency - 0.5 * channel_width
        high_freq = frequency + 0.5 * channel_width
        low_rel = (low_freq - spectral_window.min_frequency) / spectral_window.bandwidth
        high_rel = (high_freq - spectral_window.min_frequency) / spectral_window.bandwidth
        # Add an axis that will broadcast with the ranges
        low_rel = low_rel[..., np.newaxis]
        high_rel = high_rel[..., np.newaxis]
        mask = (high_rel >= self.ranges['min_fraction']) & (low_rel <= self.ranges['max_fraction'])
        return np.any(mask, axis=-1)

    @classmethod
    def from_hdf5(cls: Type[_B], hdf5: h5py.File) -> _B:
        dataset_name = 'ranges'
        data = models.get_hdf5_dataset(hdf5, dataset_name)
        expected_dtype = np.dtype([
            ('min_fraction', 'f8'),
            ('max_fraction', 'f8'),
        ])
        data = models.require_columns(f'{dataset_name} dataset', data, expected_dtype, 1)
        ranges = astropy.table.Table(data[...], copy=False)
        ranges['min_fraction'].unit = u.dimensionless_unscaled
        ranges['max_fraction'].unit = u.dimensionless_unscaled
        return cls(ranges)

    def to_hdf5(self, hdf5: h5py.File) -> None:
        hdf5.create_dataset('ranges', data=self.ranges.as_array(), track_times=False)



################################################################################
# Copyright (c) 2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Utilities for other tests."""

import os
import pathlib
import urllib.parse
from typing import ClassVar, Any
from typing_extensions import Literal

import h5py

from katsdpmodels import models


BASE_URL = 'http://test.invalid/data/'
DATA_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')


def get_data(filename: str) -> bytes:
    with open(os.path.join(DATA_DIR, filename), 'rb') as f:
        return f.read()


def get_data_url(filename: str) -> str:
    return urllib.parse.urljoin(BASE_URL, filename)


def get_file_url(filename: str) -> str:
    path = os.path.join(DATA_DIR, filename)
    return pathlib.PurePath(path).as_uri()


class DummyModel(models.SimpleHDF5Model):
    model_type: ClassVar[Literal['rfi_mask']] = 'rfi_mask'
    model_format: ClassVar[Literal['ranges']] = 'ranges'

    def __init__(self, ranges: Any) -> None:
        self.ranges = ranges
        self.is_closed = False

    @classmethod
    def from_hdf5(cls, hdf5: h5py.File) -> 'DummyModel':
        return cls(hdf5['/ranges'][:])

    def to_hdf5(self, hdf5: h5py.File) -> None:
        hdf5.create_dataset('ranges', data=self.ranges, track_times=False)

    def close(self) -> None:
        super().close()
        self.is_closed = True



################################################################################
# Copyright (c) 2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Base functionality common to all model types."""

from abc import ABC, abstractmethod
from datetime import datetime, timezone
import io
import urllib.parse
import pathlib
from typing import Mapping, BinaryIO, Optional, Any, ClassVar, Type, TypeVar, Union, overload
from typing_extensions import Literal

import h5py
import numpy as np
import numpy.lib.recfunctions
import strict_rfc3339


_E = TypeVar('_E', bound='ModelError')
_M = TypeVar('_M', bound='Model')
_H = TypeVar('_H', bound='SimpleHDF5Model')
_T = TypeVar('_T')
# typeshed doesn't correctly indicate that io.BytesIO and typing.BinaryIO
# inherit from io.IOBase, so we use this type alias for file-like parameters.
_FileLike = Union[io.IOBase, io.BytesIO, BinaryIO]


class ModelError(ValueError):
    """A model was found, but the content was incorrect."""

    def __init__(self, *args) -> None:
        super().__init__(*args)
        self.url: Optional[str] = None
        self.original_url: Optional[str] = None

    @classmethod
    def with_urls(cls: Type[_E], *args,
                  url: Optional[str] = None, original_url: Optional[str] = None) -> _E:
        exc = cls(args)
        exc.url = url
        exc.original_url = original_url
        return exc


class FileTypeError(ModelError):
    """The file type (as determined by Content-Type or extension) was not recognised."""


class ModelVersionError(ModelError):
    """The ``model_version`` attribute was missing or of the wrong type."""


class ModelTypeError(ModelError):
    """The ``model_type`` attribute was missing or did not match the expected value."""


class ModelFormatError(ModelError):
    """The ``model_format`` attribute was missing or did not match a known value."""


class DataError(ModelError):
    """The model was missing some data or it had the wrong format."""


class ChecksumError(DataError):
    """The model did not match the checksum embedded in the filename."""


class TooManyAliasesError(ModelError):
    """The limit on the number of alias redirections was reached."""


class AbsoluteAliasError(ModelError):
    """An alias retried to redirect to an absolute URL."""


class TelescopeStateError(ModelError):
    """There was a problem retrieving references from telescope state."""


class Model(ABC):
    """Base class for models.

    Models can either be loaded from file-like objects or constructed directly.
    Models loaded by the fetcher store the checksum of the raw data, and are
    considered equal if the checksums match. Otherwise, equality is by object
    identity.

    Models loaded by the fetcher should not be modified, as they may be shared
    by other users. Instead, make a copy and modify that.

    Subclassing should generally be done in two layers:

    1. A class that defines `model_type` and defines the interface for that
       model type. This will be passed to fetchers to indicate what
       model type is expected. Due to limitations in mypy, this should not
       use ``@abstractmethod`` for the interface methods.
    2. A concrete implementation that defines `model_format`.
    """

    model_type: ClassVar[str]
    model_format: ClassVar[str]
    checksum: Optional[str] = None
    target: Optional[str] = None
    comment: Optional[str] = None
    author: Optional[str] = None
    # It's required in loaded files, but optional here to allow models to be
    # built programmatically.
    version: Optional[int] = None
    created: Optional[datetime] = None

    @classmethod
    @abstractmethod
    def from_file(cls: Type[_M], file: _FileLike, url: str, *,
                  content_type: Optional[str] = None) -> _M:
        """Load a model from raw data.

        On success, the callee takes responsibility for closing `file`, either
        within the function itself or the :meth:`close` method of the returned
        model.

        If `content_type` is given, it should be used to determine the file
        type; otherwise `url` may be used instead.
        """

    @abstractmethod
    def to_file(self, file: Union[str, pathlib.Path, _FileLike], *,
                content_type: Optional[str] = None) -> None:
        """Write a model to file, overwriting any existing file.

        If `content_type` is given, it should be used to determine the file
        type; otherwise the filename may be used.
        """

    def close(self) -> None:
        """Close external resources associated with the model.

        Attempting to use the model after that results in undefined behavior.
        However, it is legal to call :meth:`close` multiple times.

        Models also implement the context manager protocol.
        """

    def __enter__(self: _M) -> _M:
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.close()

    def __eq__(self, other: object) -> Any:
        if not isinstance(other, Model):
            return NotImplemented
        elif self.checksum is not None and other.checksum is not None:
            return self.checksum == other.checksum and type(self) is type(other)
        else:
            return self is other

    def __hash__(self) -> int:
        if self.checksum is not None:
            return hash(self.checksum)
        else:
            return super().__hash__()


class SimpleHDF5Model(Model):
    """Helper base class for models that load data from HDF5.

    It does not handle lazy loading: the :meth:`from_hdf5` class method must
    load all the data out of the HDF5 file as it will be closed by the caller.
    The implementation of :meth:`from_hdf5` does not need to pull out the
    generic metadata (comment, target, author etc).
    """

    @classmethod
    def from_file(cls: Type[_H], file: _FileLike, url: str, *,
                  content_type: Optional[str] = None) -> _H:
        with file:
            if content_type is not None:
                if content_type != 'application/x-hdf5':
                    raise FileTypeError(f'Expected application/x-hdf5, not {content_type}')
            else:
                parts = urllib.parse.urlsplit(url)
                path = urllib.parse.unquote(parts.path)
                if not path.endswith(('.h5', '.hdf5')):
                    raise FileTypeError(f'Filename extension not recognised in {url}')
            with h5py.File(file, 'r') as hdf5:
                model_type = get_hdf5_attr(hdf5.attrs, 'model_type', str)
                if model_type != cls.model_type:
                    raise ModelTypeError(
                        f'Expected a model of type {cls.model_type!r}, not {model_type!r}')
                model = cls.from_hdf5(hdf5)
                try:
                    model.comment = get_hdf5_attr(hdf5.attrs, 'model_comment', str)
                    model.target = get_hdf5_attr(hdf5.attrs, 'model_target', str)
                    model.author = get_hdf5_attr(hdf5.attrs, 'model_author', str)
                    try:
                        model.version = get_hdf5_attr(hdf5.attrs, 'model_version', int,
                                                      required=True)
                    except (KeyError, TypeError) as exc:
                        raise ModelVersionError(str(exc)) from exc
                    created = get_hdf5_attr(hdf5.attrs, 'model_created', str)
                    if created is not None:
                        try:
                            model.created = rfc3339_to_datetime(created)
                        except ValueError:
                            raise DataError(f'Invalid creation timestamp {created!r}') from None
                except Exception:
                    model.close()
                    raise
                else:
                    return model

    def to_file(self, file: Union[str, pathlib.Path, _FileLike], *,
                content_type: Optional[str] = None) -> None:
        if self.version is None:
            raise ValueError('Version must be set before writing file')
        if content_type is not None:
            if content_type != 'application/x-hdf5':
                raise FileTypeError(f'Expected application/x-hdf5, not {content_type}')
        else:
            if isinstance(file, (str, pathlib.Path)):
                path = pathlib.Path(file)
            else:
                path = pathlib.Path(file.name)   # type: ignore
            if path.suffix not in {'.h5', '.hdf5'}:
                raise FileTypeError(
                    f'Expected extension of .h5 or .hdf5, not {path.suffix} '
                    '(use content_type to override if necessary)'
                )
        if isinstance(file, pathlib.Path):
            hdf5 = h5py.File(str(file), 'w')
        else:
            hdf5 = h5py.File(file, 'w')
        with hdf5:
            hdf5.attrs['model_version'] = self.version
            hdf5.attrs['model_type'] = self.model_type
            hdf5.attrs['model_format'] = self.model_format
            if self.comment is not None:
                hdf5.attrs['model_comment'] = self.comment
            if self.author is not None:
                hdf5.attrs['model_author'] = self.author
            if self.target is not None:
                hdf5.attrs['model_target'] = self.target
            if self.created is not None:
                hdf5.attrs['model_created'] = strict_rfc3339.timestamp_to_rfc3339_utcoffset(
                    self.created.timestamp()
                )
            self.to_hdf5(hdf5)

    @classmethod
    @abstractmethod
    def from_hdf5(cls: Type[_H], hdf5: h5py.File) -> _H:
        """Load a model from an HDF5 file.

        Subclasses will implement this method, but it is not intended to be
        used directly.
        """

    @abstractmethod
    def to_hdf5(self, hdf5: h5py.File) -> None:
        """Write a model to an HDF5 file.

        Subclasses will implement this method, but it is not intended to be
        used directly.
        """


@overload
def get_hdf5_attr(attrs: Mapping[str, object], name: str, required_type: Type[_T], *,
                  required: Literal[True]) -> _T: ...


@overload
def get_hdf5_attr(attrs: Mapping[str, object], name: str, required_type: Type[_T], *,
                  required: bool = False) -> Optional[_T]: ...


def get_hdf5_attr(attrs, name, required_type, *, required=False):
    """Retrieve an attribute from an HDF5 object and verify its type.

    Pass the ``attrs`` attribute of the HDF5 file, group or dataset as the
    `attrs` parameter. If the `name` is not present, returns ``None``, unless
    ``required=True`` is passed.

    The implementation includes a workaround for
    https://github.com/h5py/h5py/issues/379, which will decode byte attributes
    to Unicode. It also turns numpy scalars into plain Python types.

    Raises
    ------
    KeyError
        if `name` is not present and `required` is true.
    TypeError
        if `name` is present but is not of type `type`.
    UnicodeDecodeError
        if the attribute is :class:`bytes` that are not valid UTF-8 (only if
        `type` is :class:`str`).
    """
    try:
        value = attrs[name]
    except KeyError:
        if required:
            # The original message from h5py is less readable
            raise KeyError(f'attribute {name!r} is missing') from None
        else:
            return None
    if isinstance(value, np.generic):
        value = value.item()
    actual_type = type(value)
    if actual_type == required_type:
        return value
    elif required_type == str and isinstance(value, bytes):
        return value.decode('utf-8')
    else:
        raise TypeError(f'Expected {required_type} for {name!r}, received {actual_type}')


def get_hdf5_dataset(group: h5py.Group, name: str) -> h5py.Dataset:
    """Get a dataset from an HDF5 file, raising an exception if missing.

    The advantage of this method over directly indexing the group is that
    it will also raise the exception if a group is found instead of a
    dataset.
    """
    try:
        data = group[name]
        if isinstance(data, h5py.Group):
            raise KeyError        # It should be a dataset, not a group
    except KeyError:
        raise DataError(f'Model is missing {name} dataset') from None
    return data


def rfc3339_to_datetime(timestamp: str) -> datetime:
    """Convert a string in RFC 3339 format to a timezone-aware datetime object.

    The original timezone in the string is lost, and the returned datetime is
    based on UTC.
    """
    unix_time = strict_rfc3339.rfc3339_to_timestamp(timestamp)
    return datetime.fromtimestamp(unix_time, timezone.utc)


def require_columns(name: str, array: Any, dtype: np.dtype, ndim: int) -> Any:
    """Validate the columns in a table and the dimensionality.

    The `dtype` is the expected dtype, which must be a structured dtype. The
    array is checked for compatibility: it must have all the required fields
    (but may have more), and they must be castable to the the expected dtype.

    The return value is the input array restricted to the required columns
    and cast to the required dtype. It may be either a view or a copy,
    depending on whether any casting was required.

    This function is not designed to support nested structuring, and will not
    recursively filter out unwanted sub-structures.

    Raises
    ------
    DataError
        if the types are not compatible or the wrong number of dimensions are present.
    """
    if array.ndim != ndim:
        raise DataError(f'{name} should be {ndim}-dimensional, but is {array.ndim}-dimensional')
    if array.dtype == dtype:
        return np.asanyarray(array)
    if array.dtype.names is None:
        raise DataError(f'{name} does not have named columns')
    for col_name in dtype.names:
        if col_name not in array.dtype.names:
            raise DataError(f'{name} column {col_name} is missing')
        if not np.can_cast(array.dtype[col_name], dtype[col_name], 'same_kind'):
            raise DataError(
                f'{name} column {col_name} has type {array.dtype[col_name]}, '
                f'expected {dtype[col_name]}'
            )
    out = np.empty(array.shape, dtype)
    np.lib.recfunctions.assign_fields_by_name(out, array)
    return out



################################################################################
# Copyright (c) 2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Fetch models over HTTP."""

import errno
import functools
import io
import re
import logging
import os
import sys
import urllib.parse
from typing import (
    List, Generator, Dict, MutableMapping, Optional, Type, Callable,
    TypeVar, Any, TYPE_CHECKING
)

import requests
import urllib3

from .. import models, fetch

if TYPE_CHECKING:
    import katsdptelstate


_logger = logging.getLogger(__name__)
_T = TypeVar('_T')
_M = TypeVar('_M', bound=models.Model)
_F = TypeVar('_F', bound='Fetcher')
_TF = TypeVar('_TF', bound='TelescopeStateFetcher')
_Req = TypeVar('_Req')
_Resp = TypeVar('_Resp')

DEFAULT_RETRIES = urllib3.Retry(total=4, status_forcelist=range(500, 600), raise_on_status=False)


def _run_generator(gen: Generator[_Req, _Resp, _T],
                   handle_request: Callable[[_Req], _Resp]) -> _T:
    try:
        request = next(gen)      # Start it going
        while True:
            try:
                response = handle_request(request)
            except Exception:
                request = gen.throw(*sys.exc_info())
            else:
                request = gen.send(response)
    except StopIteration as exc:
        return exc.value
    finally:
        gen.close()


class HttpFile(io.RawIOBase):
    """File-like object that fetches byte ranges via HTTP.

    This requires the server to advertise support for byte-range requests and
    to provide a Content-Length. It is currently *not* robust against the
    content changing.

    Raises
    ------
    FileNotFoundError
        HTTP 404 error
    PermissionError
        HTTP 403 error
    OSError
        Other HTTP errors, or the server doesn't implement the required features.
    """

    # urllib3 by default sets this to gzip, deflate, but that won't play nice
    # with byte ranges because we need byte ranges from the original file, not
    # from the compressed representation.
    _HEADERS = {
        'Accept-Encoding': 'identity'
    }

    def __init__(self, session: requests.Session, url: str) -> None:
        self._session = session
        self._offset = 0
        with session.head(url, headers=self._HEADERS, allow_redirects=True) as resp:
            if resp.status_code == 404:
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), url)
            elif resp.status_code in {401, 403}:
                raise PermissionError(errno.EACCES, os.strerror(errno.EACCES), url)
            resp.raise_for_status()
            if resp.headers.get('Accept-Ranges', 'none') != 'bytes':
                raise OSError(None, 'Server does not accept byte ranges', url)
            if 'Content-Encoding' in resp.headers:
                raise OSError(None, 'Server provided Content-Encoding header', url)
            try:
                self._length = int(resp.headers['Content-Length'])
            except (KeyError, ValueError):
                raise OSError(None, 'Server did not provide Content-Length header', url) from None
            # TODO: consider storing ETag/Last-Modified to check for data
            # changing under us.
            self._url = resp.url
            self.content_type = resp.headers.get('Content-Type')

    @property
    def url(self) -> str:
        """Return the final URL.

        This may differ from the constructor argument if HTTP redirects occurred.
        """
        return self._url

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        if whence == io.SEEK_SET:
            new_offset = offset
        elif whence == io.SEEK_CUR:
            new_offset = self._offset + offset
        elif whence == io.SEEK_END:
            new_offset = self._length + offset
        else:
            raise ValueError(f'invalid whence ({whence}, should be 0, 1 or 2)')
        if new_offset < 0:
            raise OSError(errno.EINVAL, 'Invalid argument')
        self._offset = new_offset
        return new_offset

    def tell(self) -> int:
        return self._offset

    def readinto(self, b) -> int:
        start = min(self._length, self._offset)
        end = min(self._length, start + len(b)) - 1    # End is inclusive
        with self._session.get(
                self._url,
                headers={'Range': f'bytes={start}-{end}', **self._HEADERS},
                stream=True) as resp:
            resp.raise_for_status()
            if 'Content-Encoding' in resp.headers:
                raise OSError('Server provided Content-Encoding header')
            content_range = resp.headers.get('Content-Range', '')
            # RFC 7233 specifies the format
            match = re.fullmatch(r'bytes (\d+)-(\d+)/(?:\*|\d+)', content_range)
            if (resp.status_code != requests.codes.PARTIAL_CONTENT or not match
                    or int(match.group(1)) != start or int(match.group(2)) != end):
                # Tornado does not send partial content if the entire content
                # was requested, so allow that case.
                if start != 0 or end != self._length - 1:
                    raise OSError('Did not receive expected byte range')
            bytes_read = resp.raw.readinto(b)
            self._offset += bytes_read
            return bytes_read

    def close(self) -> None:
        super().close()


class Fetcher(fetch.FetcherBase):
    """Fetches and caches models, using the :mod:`requests` library.

    It caches every URL it fetches (ignoring any cache control headers), so it
    should not be reused over a long time.  It is best suited to fetching a
    batch of models, some of which may turn out to be aliases of each other.

    It should be closed with :meth:`close` when no longer in use. It also
    implements the context manager protocol for this purpose. This will close
    the retrieved models, so must only be done once the models are no longer in
    use.

    This class is *not* thread-safe.

    Parameters
    ----------
    session
        Interface for doing the actual HTTP requests. If not specified, an
        internal session will be created (which supports ``file://`` URLs), and
        closed when the fetcher is closed. If a custom session is provided it
        will *not* be closed (so it can be shared between multiple
        :class:`Fetcher` instances).
    model_cache
        A dictionary for caching models by URL. This is not typically needed,
        as the fetcher will use an internal cache if one is not provided, but
        allows fetchers to share a cache (but not in a thread-safe way!).
        If a custom cache is provided, then :meth:`close` will not close the
        models in it, and the caller is responsible for doing so.

    Raises
    ------
    :exc:`.ModelError`
        For any issues with the model itself
    :exc:`requests.RequestException`
        For any issues at the HTTP level
    """

    def __init__(self, *,
                 session: Optional[requests.Session] = None,
                 model_cache: Optional[MutableMapping[str, models.Model]] = None) -> None:
        super().__init__(model_cache=model_cache)
        if session is None:
            self._session = requests.Session()
            for protocol in ['https://', 'http://']:
                self._session.mount(
                    protocol, requests.adapters.HTTPAdapter(max_retries=DEFAULT_RETRIES)
                )
            self._close_session = True
        else:
            self._session = session
            self._close_session = False

    @property
    def session(self) -> requests.Session:
        return self._session

    def close(self) -> None:
        self._close()
        if self._close_session:
            self._session.close()

    def __enter__(self: _F) -> _F:
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.close()

    def _handle_request(self, request: fetch.Request, *,
                        lazy: bool = False) -> fetch.Response:
        assert request.response_type in {fetch.ResponseType.TEXT, fetch.ResponseType.FILE}
        parts = urllib.parse.urlsplit(request.url)
        if parts.scheme == 'file':
            return self._handle_file_scheme(request, lazy)
        elif request.response_type == fetch.ResponseType.TEXT:
            with self._session.get(request.url) as resp:
                resp.raise_for_status()
                return fetch.TextResponse(resp.url, resp.headers, resp.text)
        elif not lazy:
            with self._session.get(request.url) as resp:
                resp.raise_for_status()
                content = resp.content
                file = io.BytesIO(content)
                return fetch.FileResponse(
                    resp.url, resp.headers, file=file, content=content)
        else:
            fh = HttpFile(self._session, request.url)
            # TODO: make HttpFile return the full headers
            headers = requests.structures.CaseInsensitiveDict(
                {'Content-type': fh.content_type or 'application/octet-stream'}
            )
            return fetch.FileResponse(fh.url, headers, file=fh, content=None)

    def resolve(self, url: str) -> List[str]:
        """Follow a chain of aliases.

        Return a list of URLs found along the chain. The first element is the
        given URL and the final element is the resolved model.

        Raises
        ------
        .TooManyAliasesError
            If there were more than :data:`.MAX_ALIASES` aliases or a cycle was found.
        """
        return _run_generator(self._resolve(url), self._handle_request)

    def get(self, url: str, model_class: Type[_M], *,
            lazy: bool = False) -> _M:
        """Retrieve a single model.

        The caller must *not* close the retrieved model, as it is cached and
        a future request for the model would return the closed model. Instead,
        the caller must close the fetcher once it no longer needs any of the
        models.

        Parameters
        ----------
        url
            The URL to fetch.
        model_class
            The base class for the model type (do not use concrete
            format-specific classes). The returned model will be of this
            type.
        lazy
            If true, create a view of the HDF5 file that only retrieves data
            as it is needed. Whether this actually allows data to be loaded
            lazily depends on the `model_class`: some model classes will read
            all the data into memory on creation, in which case lazy loading
            may perform significantly worse (at the time of writing, no model
            classes support lazy loading).

            Lazy loading imposes some additional requirements:

            1. The session must not be closed while the model is in use, even
               if the fetcher is no longer needed.
            2. The checksum stored in the filename is not validated.
            3. If the model is already in the cache, the laziness setting is
               ignored and the cached model is returned.

        Raises
        ------
        .ModelError
            If there are high-level errors with the model.
        requests.exception.RequestException
            Any exceptions raised by the underlying session.
        """
        return _run_generator(
            self._get(url, model_class),
            functools.partial(self._handle_request, lazy=lazy)
        )


def fetch_model(url: str, model_class: Type[_M], *,
                session: Optional[requests.Session] = None) -> _M:
    """Retrieve a single model.

    This is a convenience function that should only be used when loading just a
    single model. If multiple models will be used instead, construct an
    instance of :class:`Fetcher` and use it to fetch models, as this will allow
    models that turn out to be the same to be shared.
    """
    # Custom cache so that fetcher won't close the model
    model_cache: Dict[str, models.Model] = {}
    with Fetcher(session=session, model_cache=model_cache) as fetcher:
        return fetcher.get(url, model_class)


class TelescopeStateFetcher(fetch.TelescopeStateFetcherBase['katsdptelstate.TelescopeState']):
    __doc__ = fetch.TelescopeStateFetcherBase.__doc__

    def __init__(self,
                 telstate: 'katsdptelstate.TelescopeState',
                 fetcher: Optional[Fetcher] = None) -> None:
        super().__init__(telstate)
        if fetcher is not None:
            self.fetcher = fetcher
            self._close_fetcher = False
        else:
            self.fetcher = Fetcher()
            self._close_fetcher = True

    @staticmethod
    def _handle_request(
            request: fetch.TelescopeStateRequest['katsdptelstate.TelescopeState']) -> str:
        return request.telstate[request.key]

    def get(self, key: str, model_class: Type[_M], *,
            telstate: Optional['katsdptelstate.TelescopeState'] = None,
            lazy: bool = False) -> _M:
        """Retrieve a single model.

        The semantics are the same as for :meth:`Fetcher.get`. Any problems
        with getting the keys from the telescope state will raise
        :exc:`.TelescopeStateError`.

        If `telstate` is provided, it is used instead of the constructor
        argument for fetching the model key; but the constructor telstate is
        still used to fetch ``sdp_model_base_url``.
        """
        url = _run_generator(self._get_url(key, telstate=telstate), self._handle_request)
        return self.fetcher.get(url, model_class, lazy=lazy)

    def close(self) -> None:
        """Clean up resources."""
        if self._close_fetcher:
            self.fetcher.close()

    def __enter__(self: _TF) -> _TF:
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.close()



################################################################################
# Copyright (c) 2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Fetch models asynchronously over HTTP."""

import io
import sys
import urllib.parse
from typing import (
    List, Dict, Generator, Optional, MutableMapping, Type, Callable, Awaitable,
    TypeVar, Union, Any, TYPE_CHECKING
)

import aiohttp
import aiohttp_retry

from .. import models, fetch

if TYPE_CHECKING:
    import katsdptelstate.aio


_T = TypeVar('_T')
_M = TypeVar('_M', bound=models.Model)
_F = TypeVar('_F', bound='Fetcher')
_TF = TypeVar('_TF', bound='TelescopeStateFetcher')
_Req = TypeVar('_Req')
_Resp = TypeVar('_Resp')
ClientSession = Union[aiohttp.ClientSession, aiohttp_retry.RetryClient]


async def _run_generator(gen: Generator[_Req, _Resp, _T],
                         handle_request: Callable[[_Req], Awaitable[_Resp]]) -> _T:
    try:
        request = next(gen)      # Start it going
        while True:
            try:
                response = await handle_request(request)
            except Exception:
                request = gen.throw(*sys.exc_info())
            else:
                request = gen.send(response)
    except StopIteration as exc:
        return exc.value
    finally:
        gen.close()


class Fetcher(fetch.FetcherBase):
    """Fetches and caches models, using the :mod:`aiohttp` library.

    It caches every URL it fetches (ignoring any cache control headers), so it
    should not be reused over a long time.  It is best suited to fetching a
    batch of models, some of which may turn out to be aliases of each other.

    It should be closed with :meth:`close` when no longer in use. It also
    implements the context manager protocol for this purpose. This will close
    the retrieved models, so must only be done once the models are no longer in
    use.

    This class is *not* thread-safe.

    Parameters
    ----------
    session
        Interface for doing the actual HTTP requests. If not specified, an
        internal session will be created, and closed when the fetcher is
        closed. If a custom session is provided it will *not* be closed (so it
        can be shared between multiple :class:`Fetcher` instances).
    model_cache
        A dictionary for caching models by URL. This is not typically needed,
        as the fetcher will use an internal cache if one is not provided, but
        allows fetchers to share a cache (but not in a thread-safe way!).
        If a custom cache is provided, then :meth:`close` will not close the
        models in it, and the caller is responsible for doing so.

    Raises
    ------
    :exc:`.ModelError`
        For any issues with the model itself
    :exc:`aiohttp.ClientError`
        For any issues at the HTTP level
    """

    def __init__(self, *,
                 session: Optional[ClientSession] = None,
                 model_cache: Optional[MutableMapping[str, models.Model]] = None) -> None:
        super().__init__(model_cache=model_cache)
        if session is None:
            retry_options = aiohttp_retry.RetryOptions(
                attempts=5,
                exceptions={aiohttp.ClientError}
            )
            self._session = aiohttp_retry.RetryClient(retry_options=retry_options)
            self._close_session = True
        else:
            self._session = session
            self._close_session = False

    @property
    def session(self) -> ClientSession:
        return self._session

    async def close(self) -> None:
        self._close()
        if self._close_session:
            await self._session.close()

    async def __aenter__(self: _F) -> _F:
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await self.close()

    async def _handle_request(self, request: fetch.Request) -> fetch.Response:
        assert request.response_type in {fetch.ResponseType.TEXT, fetch.ResponseType.FILE}
        if urllib.parse.urlsplit(request.url).scheme == 'file':
            return self._handle_file_scheme(request)
        else:
            # Note: don't pass raise_for_status=True to get, because if
            # self._session is an aiohttp_retry.RetryClient it will cause it
            # to retry on all HTTP errors (including e.g. 404) instead of
            # just server errors.
            async with self._session.get(request.url) as resp:
                resp.raise_for_status()
                if request.response_type == fetch.ResponseType.TEXT:
                    text = await resp.text()
                    return fetch.TextResponse(str(resp.url), resp.headers, text)
                else:
                    content = await resp.read()
                    file = io.BytesIO(content)
                    return fetch.FileResponse(
                        str(resp.url), resp.headers, file=file, content=content)

    async def resolve(self, url: str) -> List[str]:
        """Follow a chain of aliases.

        Return a list of URLs found along the chain. The first element is the
        given URL and the final element is the resolved model.

        Raises
        ------
        .TooManyAliasesError
            If there were more than :data:`.MAX_ALIASES` aliases or a cycle was found.
        """
        return await _run_generator(self._resolve(url), self._handle_request)

    async def get(self, url: str, model_class: Type[_M]) -> _M:
        """Retrieve a single model.

        The caller must *not* close the retrieved model, as it is cached and
        a future request for the model would return the closed model. Instead,
        the caller must close the fetcher once it no longer needs any of the
        models.

        Parameters
        ----------
        url
            The URL to fetch.
        model_class
            The base class for the model type (do not use concrete
            format-specific classes). The returned model will be of this
            type.

        Raises
        ------
        .ModelError
            If there are high-level errors with the model.
        aiohttp.ClientError
            Any exceptions raised by the underlying session.
        """
        return await _run_generator(self._get(url, model_class), self._handle_request)


async def fetch_model(url: str, model_class: Type[_M], *,
                      session: Optional[ClientSession] = None) -> _M:
    """Retrieve a single model.

    This is a convenience function that should only be used when loading just a
    single model. If multiple models will be used instead, construct an
    instance of :class:`Fetcher` and use it to fetch models, as this will allow
    models that turn out to be the same to be shared.
    """
    # Custom cache so that fetcher won't close the model
    model_cache: Dict[str, models.Model] = {}
    async with Fetcher(session=session, model_cache=model_cache) as fetcher:
        return await fetcher.get(url, model_class)


class TelescopeStateFetcher(fetch.TelescopeStateFetcherBase['katsdptelstate.aio.TelescopeState']):
    __doc__ = fetch.TelescopeStateFetcherBase.__doc__

    def __init__(self,
                 telstate: 'katsdptelstate.aio.TelescopeState',
                 fetcher: Optional[Fetcher] = None) -> None:
        super().__init__(telstate)
        if fetcher is not None:
            self.fetcher = fetcher
            self._close_fetcher = False
        else:
            self.fetcher = Fetcher()
            self._close_fetcher = True

    @staticmethod
    async def _handle_request(
            request: fetch.TelescopeStateRequest['katsdptelstate.aio.TelescopeState']) -> str:
        return await request.telstate[request.key]

    async def get(self, key: str, model_class: Type[_M], *,
                  telstate: Optional['katsdptelstate.aio.TelescopeState'] = None) -> _M:
        """Retrieve a single model.

        The semantics are the same as for :meth:`Fetcher.get`. Any problems
        with getting the keys from the telescope state will raise
        :exc:`.TelescopeStateError`.

        If `telstate` is provided, it is used instead of the constructor
        argument for fetching the model key; but the constructor telstate is
        still used to fetch ``sdp_model_base_url``.
        """
        url = await _run_generator(self._get_url(key, telstate=telstate), self._handle_request)
        return await self.fetcher.get(url, model_class)

    async def close(self) -> None:
        """Clean up resources."""
        if self._close_fetcher:
            await self.fetcher.close()

    async def __aenter__(self: _TF) -> _TF:
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await self.close()



################################################################################
# Copyright (c) 2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpmodels.fetch.requests`."""

import hashlib
import io

import h5py
import pytest
import requests
import responses
import katsdptelstate

from katsdpmodels import models, fetch
import katsdpmodels.fetch.requests as fetch_requests
from test_utils import get_data, get_data_url, get_file_url, DummyModel


@pytest.fixture
def http_file(web_server):
    with requests.Session() as session:
        with fetch_requests.HttpFile(session, web_server('all_bytes.bin')) as file:
            yield file


def test_http_file_seek_tell(http_file):
    assert http_file.tell() == 0
    http_file.seek(10, io.SEEK_SET)
    assert http_file.tell() == 10
    assert http_file.read(2) == b'\x0A\x0B'
    assert http_file.tell() == 12
    http_file.seek(5)
    assert http_file.tell() == 5
    http_file.seek(5, io.SEEK_CUR)
    assert http_file.tell() == 10
    http_file.seek(-5, io.SEEK_CUR)
    assert http_file.tell() == 5
    http_file.seek(-10, io.SEEK_END)
    assert http_file.tell() == 246
    with pytest.raises(ValueError):
        http_file.seek(0, 17)
    with pytest.raises(OSError):
        http_file.seek(-10000, io.SEEK_END)


def test_http_file_close(http_file):
    assert not http_file.closed
    http_file.close()
    assert http_file.closed
    http_file.close()
    assert http_file.closed


def test_http_file_read(http_file):
    assert http_file.read(2) == b'\x00\x01'
    assert http_file.read(3) == b'\x02\x03\x04'
    http_file.seek(-2, io.SEEK_END)
    # Short read at end of file
    assert http_file.read(4) == b'\xFE\xFF'
    assert http_file.tell() == 256


def test_http_file_not_found(web_server):
    with requests.Session() as session:
        with pytest.raises(FileNotFoundError) as exc_info:
            fetch_requests.HttpFile(session, web_server('does_not_exist'))
    assert exc_info.value.filename == web_server('does_not_exist')


def test_http_file_forbidden(mock_responses):
    url = get_data_url('does_not_exist')
    mock_responses.add(responses.HEAD, url, status=403)
    with requests.Session() as session:
        with pytest.raises(PermissionError) as exc_info:
            fetch_requests.HttpFile(session, url)
    assert exc_info.value.filename == url


def test_http_file_ranges_not_accepted(mock_responses):
    url = get_data_url('rfi_mask_ranges.h5')
    with requests.Session() as session:
        with pytest.raises(OSError, match='Server does not accept byte ranges') as exc_info:
            fetch_requests.HttpFile(session, url)
    assert exc_info.value.filename == url


def test_http_file_no_content_length(mock_responses):
    url = get_data_url('rfi_mask_ranges.h5')
    mock_responses.replace(responses.HEAD, url, content_type='application/x-hdf5',
                           headers={'Accept-Ranges': 'bytes'})
    with requests.Session() as session:
        with pytest.raises(OSError,
                           match='Server did not provide Content-Length header') as exc_info:
            fetch_requests.HttpFile(session, url)
    assert exc_info.value.filename == url


def test_http_file_content_encoding(mock_responses):
    url = get_data_url('rfi_mask_ranges.h5')
    mock_responses.replace(
        responses.HEAD, url,
        headers={
            'Content-Length': str(len(get_data('rfi_mask_ranges.h5'))),
            'Accept-Ranges': 'bytes',
            'Content-Encoding': 'gzip'
        }
    )
    with requests.Session() as session:
        with pytest.raises(OSError,
                           match='Server provided Content-Encoding header') as exc_info:
            fetch_requests.HttpFile(session, url)
    assert exc_info.value.filename == url


def test_http_file_content_encoding_get(mock_responses):
    url = get_data_url('rfi_mask_ranges.h5')
    mock_responses.replace(
        responses.HEAD, url,
        headers={
            'Content-Length': str(len(get_data('rfi_mask_ranges.h5'))),
            'Accept-Ranges': 'bytes'
        }
    )
    mock_responses.replace(
        responses.GET, url,
        headers={
            'Content-Length': str(len(get_data('rfi_mask_ranges.h5'))),
            'Accept-Ranges': 'bytes',
            'Content-Encoding': 'gzip'
        }
    )
    with requests.Session() as session:
        with fetch_requests.HttpFile(session, url) as file:
            with pytest.raises(OSError, match='Server provided Content-Encoding header'):
                file.read()


def test_http_file_url_attr(mock_responses, web_server):
    url = get_data_url('subdir/redirect.h5')
    new_url = web_server('rfi_mask_ranges.h5')
    mock_responses.add(
        responses.HEAD, url,
        headers={'Location': new_url},
        status=307
    )
    mock_responses.add_passthru(web_server(''))
    with requests.Session() as session:
        with fetch_requests.HttpFile(session, url) as file:
            assert file.url == new_url


def test_http_file_range_ignored(mock_responses):
    url = get_data_url('rfi_mask_ranges.h5')
    data = get_data('rfi_mask_ranges.h5')
    mock_responses.replace(
        responses.HEAD, url,
        headers={
            'Accept-Ranges': 'bytes',
            'Content-Length': str(len(data))
        }
    )
    mock_responses.replace(responses.GET, url, body=data, stream=True)
    with requests.Session() as session:
        with fetch_requests.HttpFile(session, url) as file:
            with pytest.raises(OSError, match='Did not receive expected byte range'):
                file.read(10)
            # Reading the whole file should work even if the server doesn't send
            # back partial content.
            file.seek(0)
            test_data = file.read(len(data))
            assert test_data == data


@pytest.mark.parametrize('use_file', [True, False])
@pytest.mark.parametrize('filename', ['rfi_mask_ranges.h5', 'direct.alias', 'indirect.alias'])
def test_fetch_model_simple(use_file, filename, web_server) -> None:
    url = get_file_url(filename) if use_file else web_server(filename)
    with fetch_requests.fetch_model(url, DummyModel) as model:
        assert len(model.ranges) == 2
        assert not model.is_closed
    assert model.is_closed


def test_fetch_model_retry(web_server) -> None:
    """Test that retry happens on a 5xx server error."""
    url = web_server('rfi_mask_ranges.h5').replace('/static/', '/failonce/')
    with fetch_requests.fetch_model(url, DummyModel) as model:
        assert len(model.ranges) == 2
        assert not model.is_closed
    assert model.is_closed


def test_fetch_model_alias_loop(web_server) -> None:
    url = web_server('loop.alias')
    with pytest.raises(models.TooManyAliasesError) as exc_info:
        fetch_requests.fetch_model(url, DummyModel)
    assert exc_info.value.url == url
    assert exc_info.value.original_url == url


def test_fetch_model_too_many_aliases(web_server, monkeypatch) -> None:
    monkeypatch.setattr(fetch, 'MAX_ALIASES', 1)
    with pytest.raises(models.TooManyAliasesError):
        fetch_requests.fetch_model(web_server('indirect.alias'), DummyModel)
    # Check that 1 level of alias is still permitted
    with fetch_requests.fetch_model(web_server('direct.alias'), DummyModel):
        pass


def test_fetch_model_absolute_alias(web_server) -> None:
    url = web_server('to_file.alias')
    with pytest.raises(models.AbsoluteAliasError) as exc_info:
        fetch_requests.fetch_model(url, DummyModel)
    assert exc_info.value.url == url
    assert exc_info.value.original_url == url


@pytest.mark.parametrize('filename', ['bad_model_type.h5', 'no_model_type.h5'])
def test_fetch_model_model_type_error(filename, web_server) -> None:
    url = web_server(filename)
    with pytest.raises(models.ModelTypeError, match='rfi_mask') as exc_info:
        fetch_requests.fetch_model(url, DummyModel)
    assert exc_info.value.url == url
    assert exc_info.value.original_url == url


def test_fetch_model_bad_created(web_server) -> None:
    url = web_server('bad_created.h5')
    with pytest.raises(models.DataError, match='Invalid creation timestamp') as exc_info:
        fetch_requests.fetch_model(url, DummyModel)
    assert exc_info.value.url == url
    assert exc_info.value.original_url == url


@pytest.mark.parametrize('filename', ['bad_model_version.h5', 'no_model_version.h5'])
def test_fetch_model_model_version_error(filename, web_server) -> None:
    url = web_server(filename)
    with pytest.raises(models.ModelVersionError) as exc_info:
        fetch_requests.fetch_model(url, DummyModel)
    assert exc_info.value.url == url
    assert exc_info.value.original_url == url
    assert 'model_version' in str(exc_info.value)


def test_fetch_model_cached_model_type_error(web_server) -> None:
    class OtherModel(models.SimpleHDF5Model):
        model_type = 'other'

        @classmethod
        def from_hdf5(cls, hdf5: h5py.File) -> 'OtherModel':
            return cls()

        def to_hdf5(self, hdf5: h5py.File) -> None:
            pass

    url = web_server('rfi_mask_ranges.h5')
    with fetch_requests.Fetcher() as fetcher:
        fetcher.get(url, DummyModel)
        with pytest.raises(models.ModelTypeError) as exc_info:
            fetcher.get(url, OtherModel)
        assert exc_info.value.url == url
        assert exc_info.value.original_url == url


@pytest.mark.parametrize('lazy', [True, False])
def test_fetch_model_bad_content_type(mock_responses, lazy) -> None:
    data = get_data('rfi_mask_ranges.h5')
    url = get_data_url('bad_content_type.h5')
    mock_responses.add(
        responses.HEAD, url, content_type='image/png',
        headers={
            'Accept-Ranges': 'bytes',
            'Content-Length': str(len(data))
        }
    )
    mock_responses.add(responses.GET, url, content_type='image/png', body=data)
    with pytest.raises(models.FileTypeError,
                       match='Expected application/x-hdf5, not image/png') as exc_info:
        with fetch_requests.Fetcher() as fetcher:
            fetcher.get(url, DummyModel, lazy=lazy)
    assert exc_info.value.url == url
    assert exc_info.value.original_url == url


@pytest.mark.parametrize('lazy', [True, False])
def test_fetch_model_bad_extension(web_server, lazy) -> None:
    url = web_server('wrong_extension.blah')
    with pytest.raises(models.FileTypeError) as exc_info:
        with fetch_requests.Fetcher() as fetcher:
            fetcher.get(url, DummyModel, lazy=lazy)
    assert exc_info.value.url == url
    assert exc_info.value.original_url == url


def test_fetch_model_not_hdf5(web_server) -> None:
    url = web_server('not_hdf5.h5')
    with pytest.raises(models.DataError) as exc_info:
        fetch_requests.fetch_model(url, DummyModel)
    assert exc_info.value.url == url
    assert exc_info.value.original_url == url


def test_fetch_model_checksum_ok(mock_responses) -> None:
    data = get_data('rfi_mask_ranges.h5')
    digest = hashlib.sha256(data).hexdigest()
    url = get_data_url(f'sha256_{digest}.h5')
    mock_responses.add(responses.GET, url, content_type='application/x-hdf5', body=data)
    with fetch_requests.fetch_model(url, DummyModel) as model:
        assert model.checksum == digest


def test_fetch_model_checksum_bad(mock_responses) -> None:
    data = get_data('rfi_mask_ranges.h5')
    digest = hashlib.sha256(data).hexdigest()
    url = get_data_url(f'sha256_{digest}.h5')
    # Now invalidate the digest
    data += b'blahblahblah'
    mock_responses.add(responses.GET, url, content_type='application/x-hdf5', body=data)
    with pytest.raises(models.ChecksumError) as exc_info:
        fetch_requests.fetch_model(url, DummyModel)
    assert exc_info.value.url == url


@pytest.mark.parametrize('filename', ['does_not_exist.h5', 'does_not_exist.alias'])
def test_fetch_model_bad_http_status(filename, web_server) -> None:
    url = web_server(filename)
    with pytest.raises(requests.HTTPError) as exc_info:
        fetch_requests.fetch_model(url, DummyModel)
    assert exc_info.value.response.status_code == 404


def test_fetch_model_http_redirect(mock_responses) -> None:
    url = get_data_url('subdir/redirect.alias')
    mock_responses.add(responses.GET, url, headers={'Location': '../direct.alias'}, status=307)
    with fetch_requests.fetch_model(url, DummyModel) as model:
        assert len(model.ranges) == 2


def test_fetch_model_connection_error(mock_responses) -> None:
    # responses raises ConnectionError for any unregistered URL
    with pytest.raises(requests.ConnectionError):
        fetch_requests.fetch_model(get_data_url('does_not_exist.h5'), DummyModel)


def test_fetcher_caching(mock_responses) -> None:
    with fetch_requests.Fetcher() as fetcher:
        model1 = fetcher.get(get_data_url('rfi_mask_ranges.h5'), DummyModel)
        model2 = fetcher.get(get_data_url('indirect.alias'), DummyModel)
        model3 = fetcher.get(get_data_url('direct.alias'), DummyModel)
        assert model1 is model2
        assert model1 is model3
        assert not model1.is_closed
    assert len(mock_responses.calls) == 3
    assert model1.is_closed


class DummySession(requests.Session):
    def __init__(self) -> None:
        super().__init__()
        self.closed = False
        self.calls = 0

    def request(self, method: str, url: str, **kwargs) -> requests.Response:   # type: ignore
        assert not self.closed
        self.calls += 1
        return super().request(method, url, **kwargs)

    def close(self) -> None:
        super().close()
        self.closed = True


def test_fetcher_custom_session(web_server) -> None:
    session = DummySession()
    with session:
        with fetch_requests.Fetcher(session=session) as fetcher:
            assert fetcher.session is session
            fetcher.get(web_server('direct.alias'), DummyModel)
        assert session.calls == 2
        assert not session.closed
    assert session.closed


def test_custom_session(web_server) -> None:
    session = DummySession()
    with session:
        fetch_requests.fetch_model(web_server('direct.alias'), DummyModel, session=session)
        assert session.calls == 2
        assert not session.closed
    assert session.closed


def test_fetcher_resolve(web_server) -> None:
    url = web_server('indirect.alias')
    with fetch_requests.Fetcher() as fetcher:
        urls = fetcher.resolve(url)
    assert urls == [
        web_server('indirect.alias'),
        web_server('direct.alias'),
        web_server('rfi_mask_ranges.h5')
    ]


def test_lazy(web_server) -> None:
    with fetch_requests.Fetcher() as fetcher:
        model = fetcher.get(web_server('rfi_mask_ranges.h5'), DummyModel, lazy=True)
        assert len(model.ranges) == 2
        assert not model.is_closed
    assert model.is_closed


def test_lazy_local() -> None:
    with fetch_requests.Fetcher() as fetcher:
        model = fetcher.get(get_file_url('rfi_mask_ranges.h5'), DummyModel, lazy=True)
        assert len(model.ranges) == 2


@pytest.fixture
def telstate_fetcher(telstate):
    fetcher = fetch_requests.TelescopeStateFetcher(telstate)
    with fetcher:
        yield fetcher


def test_telescope_state_fetcher_missing_base_url(telstate) -> None:
    telstate.delete('sdp_model_base_url')
    with fetch_requests.TelescopeStateFetcher(telstate) as fetcher:
        with pytest.raises(models.TelescopeStateError, match='not found'):
            fetcher.get('model_key', DummyModel)


def test_telescope_state_fetcher_bad_base_url(telstate) -> None:
    telstate.delete('sdp_model_base_url')
    telstate['sdp_model_base_url'] = b'Not a string'
    with fetch_requests.TelescopeStateFetcher(telstate) as fetcher:
        with pytest.raises(models.TelescopeStateError, match='should be a str'):
            fetcher.get('model_key', DummyModel)


def test_telescope_state_fetcher_missing_key(telstate_fetcher) -> None:
    with pytest.raises(models.TelescopeStateError, match='not found'):
        telstate_fetcher.get('missing_key', DummyModel)


def test_telescope_state_fetcher_bad_key(telstate, telstate_fetcher) -> None:
    telstate['bad_key'] = 123
    with pytest.raises(models.TelescopeStateError, match='should be a str'):
        telstate_fetcher.get('bad_key', DummyModel)


def test_telescope_state_fetcher_connection_error(telstate_fetcher, mocker) -> None:
    mocker.patch('katsdptelstate.TelescopeState.__getitem__',
                 side_effect=katsdptelstate.ConnectionError('test error'))
    with pytest.raises(models.TelescopeStateError, match='test error'):
        telstate_fetcher.get('model_key', DummyModel)


def test_telescope_state_fetcher_good(telstate_fetcher, mock_responses) -> None:
    model = telstate_fetcher.get('model_key', DummyModel)
    assert len(model.ranges) == 2


def test_telescope_state_fetcher_override(telstate_fetcher, mock_responses) -> None:
    telstate2 = katsdptelstate.TelescopeState()
    telstate2['another_key'] = 'rfi_mask_ranges.h5'
    model = telstate_fetcher.get('another_key', DummyModel, telstate=telstate2)
    assert len(model.ranges) == 2



################################################################################
# Copyright (c) 2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpmodels.rfi_mask`"""

import io
from typing import Generator

import astropy.units as u
import astropy.table
import numpy as np
import pytest

from katsdpmodels import models, rfi_mask
import katsdpmodels.fetch.requests as fetch_requests
from test_utils import get_data_url


@pytest.fixture
def ranges_model(web_server) -> Generator[rfi_mask.RFIMask, None, None]:
    with fetch_requests.fetch_model(web_server('rfi_mask_ranges.h5'), rfi_mask.RFIMask) as model:
        yield model


@pytest.mark.parametrize(
    'frequency,baseline_length,result',
    [
        (90e6 * u.Hz, 2 * u.m, False),
        (110e6 * u.Hz, 2 * u.m, True),
        (110 * u.MHz, 2 * u.m, True),
        (110e6 * u.Hz, 2000 * u.m, False),
        (100 * u.MHz, 2 * u.km, False),
        (600e6 * u.Hz, 2000 * u.m, True),
        (600 * u.MHz, 2 * u.km, True),
        (600 * u.MHz, 0 * u.m, False)
    ])
def test_is_masked_scalar(frequency: u.Quantity, baseline_length: u.Quantity, result: bool,
                          ranges_model: rfi_mask.RFIMask) -> None:
    assert ranges_model.is_masked(frequency, baseline_length) == result


def test_is_masked_vector(ranges_model: rfi_mask.RFIMaskRanges):
    frequency = u.Quantity([90, 110, 90, 110, 300, 600, 600, 600], u.MHz)
    baseline_length = u.Quantity([2, 2, 2000, 2000, 2, 2, 2000, 0], u.m)
    result = ranges_model.is_masked(frequency, baseline_length)
    expected = np.array([False, True, False, False, False, True, True, False])
    np.testing.assert_array_equal(result, expected)
    # Test broadcasting of inputs against each other
    result = ranges_model.is_masked(frequency, 2 * u.m)
    expected = np.array([False, True, False, True, False, True, True, True])
    np.testing.assert_array_equal(result, expected)


def test_is_masked_channel_bandwidth(ranges_model: rfi_mask.RFIMaskRanges):
    frequency = u.Quantity([70, 90, 110, 190, 210, 230], u.MHz)
    result = ranges_model.is_masked(frequency, 1 * u.m, 20 * u.MHz)
    expected = np.array([False, True, True, True, True, False])
    np.testing.assert_array_equal(result, expected)


def test_is_masked_auto_correlations(ranges_model: rfi_mask.RFIMaskRanges) -> None:
    ranges_model = rfi_mask.RFIMaskRanges(ranges_model.ranges, True)
    assert ranges_model.is_masked(600 * u.MHz, 0 * u.m)


@pytest.mark.parametrize(
    'frequency,result',
    [
        (90e6 * u.Hz, -1 * u.m),
        (90 * u.MHz, -1 * u.m),
        (110e6 * u.Hz, 1000 * u.m),
        (110 * u.MHz, 1000 * u.m),
        (300 * u.MHz, -1 * u.m),
        (600 * u.MHz, np.inf * u.m)
    ])
def test_max_baseline_length_scalar(frequency: u.Quantity, result: u.Quantity,
                                    ranges_model: rfi_mask.RFIMask) -> None:
    assert ranges_model.max_baseline_length(frequency) == result


def test_max_baseline_length_vector(ranges_model: rfi_mask.RFIMask) -> None:
    frequency = u.Quantity([90, 110, 300, 600, 900], u.MHz)
    result = ranges_model.max_baseline_length(frequency)
    np.testing.assert_array_equal(
        result.to_value(u.m),
        [-1, 1000, -1, np.inf, -1]
    )


def test_max_baseline_length_channel_width(ranges_model: rfi_mask.RFIMask) -> None:
    frequency = u.Quantity([70, 90, 110, 190, 210, 230, 470, 490], u.MHz)
    result = ranges_model.max_baseline_length(frequency, 20 * u.MHz)
    np.testing.assert_array_equal(
        result.to_value(u.m),
        [-1, 1000, 1000, 1000, 1000, -1, -1, np.inf]
    )


def test_max_baseline_length_empty(ranges_model: rfi_mask.RFIMaskRanges) -> None:
    ranges_model.ranges.remove_rows(np.s_[:])
    assert ranges_model.max_baseline_length(1 * u.Hz) == -1 * u.m
    result = ranges_model.max_baseline_length([1, 2] * u.Hz)
    np.testing.assert_array_equal(result.to_value(u.m), [-1.0, -1.0])


@pytest.mark.parametrize(
    'filename',
    ['rfi_mask_missing_dataset.h5', 'rfi_mask_ranges_is_group.h5'])
def test_missing_dataset(filename: str, mock_responses) -> None:
    url = get_data_url(filename)
    with pytest.raises(models.DataError, match='Model is missing ranges dataset'):
        fetch_requests.fetch_model(url, rfi_mask.RFIMask)


def test_bad_shape(mock_responses) -> None:
    url = get_data_url('rfi_mask_ranges_2d.h5')
    with pytest.raises(models.DataError,
                       match='ranges dataset should be 1-dimensional, but is 2-dimensional'):
        fetch_requests.fetch_model(url, rfi_mask.RFIMask)


def test_bad_model_format(mock_responses) -> None:
    url = get_data_url('rfi_mask_bad_format.h5')
    with pytest.raises(models.ModelFormatError) as exc_info:
        fetch_requests.fetch_model(url, rfi_mask.RFIMask)
    assert str(exc_info.value) == "Unknown model_format 'not_ranges' for rfi_mask"


def test_metadata(web_server) -> None:
    with fetch_requests.fetch_model(
            web_server('rfi_mask_ranges_metadata.h5'), rfi_mask.RFIMask) as model:
        assert model.comment == 'Test model'
        assert model.author == 'Test author'
        assert model.target == 'Test target'
        assert model.created is not None
        assert model.created.isoformat() == '2020-06-11T11:11:00+00:00'


def test_construct_bad_units() -> None:
    ranges = astropy.table.Table(
        [(1, 2, 3), (4, 5, 6), (7, 8, 9)],
        names=('min_frequency', 'max_frequency', 'max_baseline'),
        dtype=(np.float64, np.float64, np.float64)
    )
    # No units at all
    with pytest.raises(u.UnitConversionError):
        rfi_mask.RFIMaskRanges(ranges, True)
    # Incompatible units for baseline
    ranges['min_frequency'].unit = u.Hz
    ranges['max_frequency'].unit = u.Hz
    ranges['max_baseline'].unit = u.Hz
    with pytest.raises(u.UnitConversionError):
        rfi_mask.RFIMaskRanges(ranges, True)


def test_construct_incompatible_types() -> None:
    ranges = astropy.table.Table(
        [(1, 2, 3), (4, 5, 6), ('a', 'b', 'c')],
        names=('min_frequency', 'max_frequency', 'max_baseline'),
        dtype=(np.float64, np.float64, 'S1')
    )
    with pytest.raises(ValueError):
        rfi_mask.RFIMaskRanges(ranges, True)


def test_construct_missing_column() -> None:
    ranges = astropy.table.Table(
        [(1, 2, 3), (4, 5, 6)],
        names=('min_frequency', 'max_frequency'),
        dtype=(np.float64, np.float64)
    )
    with pytest.raises(KeyError):
        rfi_mask.RFIMaskRanges(ranges, True)


def test_to_file(ranges_model):
    fh = io.BytesIO()
    ranges_model.to_file(fh, content_type='application/x-hdf5')
    fh.seek(0)
    new_model = rfi_mask.RFIMask.from_file(fh, 'http://test.invalid/test.h5',
                                           content_type='application/x-hdf5')
    np.testing.assert_array_equal(new_model.ranges, ranges_model.ranges)
    assert new_model.mask_auto_correlations == ranges_model.mask_auto_correlations



################################################################################
# Copyright (c) 2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import contextlib
import enum
import hashlib
import logging
import pathlib
import re
import urllib.parse
from typing import List, Dict, Generator, Mapping, MutableMapping, Optional, Type, TypeVar, Generic

from .. import models
from ..models import _FileLike


MAX_ALIASES = 30     #: Maximum number of aliases that will be followed to find a model
_logger = logging.getLogger(__name__)
_M = TypeVar('_M', bound=models.Model)
_TS = TypeVar('_TS')


class ResponseType(enum.Enum):
    TEXT = 0
    FILE = 1


class Request:
    """Bare essentials of an HTTP request."""

    # TODO: use attrs?
    def __init__(self, url: str, response_type: ResponseType) -> None:
        self.url = url
        self.response_type = response_type


class Response:
    """Bare essentials of a response to a :class:`Request`.

    Parameters
    ----------
    url
        The final URL of the request, after handling any redirects.
    headers
        A case-insensitive mapping of HTTP headers
    """

    def __init__(self, url: str, headers: Mapping[str, str]) -> None:
        self.url = url
        self.headers = headers

    @property
    def content_type(self) -> Optional[str]:
        """Return the MIME content type of the request.

        Any parameters (like encoding) are stripped off. If the Content-Type
        header is missing or it is application/octet-stream, returns
        ``None``.
        """
        raw = self.headers.get('Content-Type', 'application/octet-stream')
        # Split parameters like encoding
        content_type = raw.split(';')[0].strip()
        if content_type == 'application/octet-stream':
            # This is a generic/fallback content type that doesn't convey any
            # useful information.
            return None
        else:
            return content_type


class TextResponse(Response):
    """Response to a :const:`ResponseType.TEXT` request."""

    def __init__(self, url: str, headers: Mapping[str, str], text: str) -> None:
        super().__init__(url, headers)
        self.text = text


class FileResponse(Response):
    """Response to a :const:`ResponseType.FILE` request.

    Parameters
    ----------
    url
        The final URL of the request, after handling any redirects.
    headers
        A case-insensitive mapping of HTTP headers
    file
        A file-like object from which the binary content can be read. The
        receiver of the response is responsible for closing it.
    content
        If available, the content of the response. It is used only for
        verifying checksums.
    """

    def __init__(self, url: str, headers: Mapping[str, str],
                 file: _FileLike, content: Optional[bytes] = None) -> None:
        super().__init__(url, headers)
        self.file = file
        self.content = content


class FetcherBase:
    """Base class for HTTP fetcher implementations.

    It caches every URL it fetches (ignoring any cache control headers), so it
    should not be reused over a long time.  It is best suited to fetching a
    batch of models, some of which may turn out to be aliases of each other.

    It does not perform any I/O itself. Instead, it provides generators
    that yield :class:`Requests <Request>` and expects to receive
    :class:`Responses <Response>` in reply. The subclass is responsible for
    producing the responses to requests. This design allows the core logic to
    be shared between synchronous and asynchronous implementations.

    Parameters
    ----------
    model_cache
        A dictionary for caching models by URL. This is not typically needed,
        as the fetcher will use an internal cache if one is not provided, but
        allows fetchers to share a cache (but not in a thread-safe way!).
        If a custom cache is provided, then closing the fetcher will not close
        the models in it, and the caller is responsible for doing so.
    """

    def __init__(self, *, model_cache: Optional[MutableMapping[str, models.Model]] = None) -> None:
        self._alias_cache: Dict[str, str] = {}
        if model_cache is not None:
            self._model_cache = model_cache
            self._close_models = False
        else:
            self._model_cache = {}
            self._close_models = True

    def _resolve(self, url: str) -> Generator[Request, Response, List[str]]:
        chain = [url]
        parts = urllib.parse.urlsplit(url)
        while urllib.parse.unquote(parts.path).endswith('.alias'):
            if len(chain) > MAX_ALIASES:
                raise models.TooManyAliasesError.with_urls(
                    f'Reached limit of {MAX_ALIASES} levels of aliases',
                    url=url, original_url=chain[0])
            if url in self._alias_cache:
                new_url = self._alias_cache[url]
            else:
                request = Request(url, ResponseType.TEXT)
                response = yield request
                assert isinstance(response, TextResponse)
                rel_path = response.text.rstrip()
                rel_parts = urllib.parse.urlsplit(rel_path)
                if rel_parts.scheme or rel_parts.netloc:
                    raise models.AbsoluteAliasError.with_urls(
                        f'Aliases may only contain relative URLs, not {rel_path!r}',
                        url=url, original_url=chain[0]
                    )
                new_url = urllib.parse.urljoin(response.url, rel_path)
                self._alias_cache[url] = new_url
            if new_url in chain:
                raise models.TooManyAliasesError.with_urls(
                    f'Cycle detected starting from {new_url}',
                    url=new_url, original_url=chain[0])
            chain.append(new_url)
            _logger.debug('Redirecting from %s to %s', url, new_url)
            url = new_url
            parts = urllib.parse.urlsplit(url)
        return chain

    def _handle_file_scheme(self, request: Request, lazy: bool = False) -> Response:
        """Handle a request with ``file://`` scheme.

        Subclasses can delegate to this function to deal with such URLs.
        """
        assert request.response_type in {ResponseType.TEXT, ResponseType.FILE}
        parts = urllib.parse.urlsplit(request.url)
        path = pathlib.Path(urllib.parse.unquote(parts.path)).resolve()
        response: Response
        if request.response_type == ResponseType.TEXT:
            with open(path, 'r', errors='replace') as f:
                text = f.read()
                response = TextResponse(path.as_uri(), {}, text)
        else:
            with contextlib.ExitStack() as exit_stack:
                file = open(path, 'rb')
                exit_stack.callback(file.close)
                if not lazy and file.seekable:
                    content: Optional[bytes] = file.read()
                    file.seek(0)
                else:
                    content = None
                exit_stack.pop_all()
                response = FileResponse(path.as_uri(), {}, file, content)
        return response

    def _get(self, url: str, model_class: Type[_M]) -> Generator[Request, Response, _M]:
        original_url = url
        url = (yield from self._resolve(url))[-1]
        if url in self._model_cache:
            model = self._model_cache[url]
            if model_class.model_type != model.model_type:
                raise models.ModelTypeError.with_urls(
                    f'Expected a model of type {model_class.model_type!r}, '
                    f'not {model.model_type!r}',
                    url=url, original_url=original_url)
            if not isinstance(model, model_class):
                raise TypeError('model_class should be the base class for the model type')
            return model

        with contextlib.ExitStack() as exit_stack:
            checksum: Optional[str] = None
            try:
                request = Request(url, ResponseType.FILE)
                response = yield request
                assert isinstance(response, FileResponse)
                exit_stack.callback(response.file.close)
                url = response.url
                if response.content is not None:
                    checksum = hashlib.sha256(response.content).hexdigest()
                    parts = urllib.parse.urlsplit(url)
                    match = re.search(r'/sha256_([a-z0-9]+)\.[^/]+$',
                                      urllib.parse.unquote(parts.path))
                    if match and checksum != match.group(1):
                        raise models.ChecksumError(f'Content at {url} did not match checksum')
                try:
                    new_model = model_class.from_file(response.file, url,
                                                      content_type=response.content_type)
                except models.ModelError:
                    raise
                except Exception as exc:
                    raise models.DataError(f'Failed to load model from {url}: {exc}') from exc
            except models.ModelError as exc:
                exc.original_url = original_url
                exc.url = url
                raise
            else:
                exit_stack.pop_all()   # new_model now owns fh, or has closed it

        new_model.checksum = checksum
        self._model_cache[url] = new_model
        return new_model

    def _close(self) -> None:
        """Release the resources associated with the fetcher.

        After this the fetcher should not be used further, except that it is
        legal to call this method multiple times.
        """
        self._alias_cache.clear()
        if self._close_models:
            for model in self._model_cache.values():
                model.close()
            self._model_cache.clear()
        else:
            self._model_cache = {}     # Allow garbage collection of the old cache


class TelescopeStateRequest(Generic[_TS]):
    """A request for a key from a telescope state."""

    def __init__(self, telstate: _TS, key: str) -> None:
        self.telstate = telstate
        self.key = key


_TSGenerator = Generator[TelescopeStateRequest[_TS], object, str]


class TelescopeStateFetcherBase(Generic[_TS]):
    """Fetches models that are referenced by telescope state.

    The telescope state must have a ``sdp_model_base_url`` key with a base
    URL, and a key per model with an URL relative to this one. If it is
    missing then a :exc:`KeyError` will be raised from :meth:`get`, rather
    than the constructor.

    If no fetcher is provided, an internal one will be created, and closed
    when this object is closed.

    This class is *not* thread-safe.
    """

    def __init__(self, telstate: _TS) -> None:
        self.telstate = telstate
        # Cache of sdp_model_base_url
        self._base_url: Optional[str] = None

    def _get_str(self, telstate: _TS, key: str) -> _TSGenerator:
        """Get a string-valued key from the telescope state.

        It yields the key to the caller, which should either respond
        with the value or throw in an exception.
        """
        try:
            value = yield TelescopeStateRequest(telstate, key)
        except Exception as exc:
            raise models.TelescopeStateError(f'could not get {key}: {exc}') from exc
        if not isinstance(value, str):
            type_ = type(value)
            raise models.TelescopeStateError(f'{key} should be a str, not {type_}')
        else:
            return value

    def _get_url(self, key: str, *, telstate: Optional[_TS] = None) -> _TSGenerator:
        if self._base_url is None:
            self._base_url = yield from self._get_str(self.telstate, 'sdp_model_base_url')
        if telstate is None:
            telstate = self.telstate
        rel_url = yield from self._get_str(telstate, key)
        return urllib.parse.urljoin(self._base_url, rel_url)

# Automatically added by katversion
__version__ = '0.1'



################################################################################
# Copyright (c) 2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Masks for radio-frequency interference."""

from typing import Type, TypeVar, ClassVar, Any
from typing_extensions import Literal

import numpy as np
import astropy.units as u
import astropy.table
import h5py

from . import models


_R = TypeVar('_R', bound='RFIMaskRanges')


class RFIMask(models.SimpleHDF5Model):
    model_type: ClassVar[Literal['rfi_mask']] = 'rfi_mask'

    # Methods are not marked @abstractmethod as it causes issues with mypy:
    # https://github.com/python/mypy/issues/4717

    def is_masked(self, frequency: u.Quantity, baseline_length: u.Quantity,
                  channel_bandwidth: u.Quantity = 0 * u.Hz) -> Any:
        """Determine whether given frequency is masked for the given baseline length.

        The return value is either a boolean (if `frequency` and
        `baseline_length` are scalar) or an array of boolean if they're arrays,
        with the usual broadcasting rules applying.

        A channel is masked if any part of the channel overlaps with RFI. The
        channel has width `channel_bandwidth` and is centred on `frequency`.
        """
        raise NotImplementedError()      # pragma: nocover

    def max_baseline_length(self, frequency: u.Quantity,
                            channel_bandwidth: u.Quantity = 0 * u.Hz) -> Any:
        """Determine maximum baseline length for which data at `frequency` should be masked.

        If the frequency is not masked at all, returns a negative length, and
        if it is masked at all baseline lengths, returns +inf. One may also
        supply an array of frequencies and receive an array of responses.
        """
        raise NotImplementedError()      # pragma: nocover

    @property
    def mask_auto_correlations(self) -> bool:
        """Return whether auto-correlations should be masked too.

        Auto-correlations are defined as baselines with zero length, which
        includes cross-hand polarization products.
        """
        raise NotImplementedError()      # pragma: nocover

    @classmethod
    def from_hdf5(cls, hdf5: h5py.File) -> 'RFIMask':
        model_format = models.get_hdf5_attr(hdf5.attrs, 'model_format', str) or ''
        if model_format == 'ranges':
            return RFIMaskRanges.from_hdf5(hdf5)
        else:
            raise models.ModelFormatError(
                f'Unknown model_format {model_format!r} for {cls.model_type}')

    def to_hdf5(self, hdf5: h5py.File) -> None:
        raise NotImplementedError()      # pragma: nocover


class RFIMaskRanges(RFIMask):
    model_format: ClassVar[Literal['ranges']] = 'ranges'

    def __init__(self, ranges: astropy.table.Table, mask_auto_correlations: bool) -> None:
        super().__init__()
        cols = ('min_frequency', 'max_frequency', 'max_baseline')
        units = (u.Hz, u.Hz, u.m)
        self.ranges = astropy.table.QTable(
            [ranges[col] for col in cols],
            names=cols,
            dtype=(np.float64, np.float64, np.float64)
        )
        # Canonicalise the units to simplify to_hdf5 (and also remove the
        # cost of conversions when methods are called with canonical units).
        for col, unit in zip(cols, units):
            # Ensure we haven't been given unit-less data, as <<= will inject
            # the unit (see https://github.com/astropy/astropy/issues/10514).
            if self.ranges[col].unit is None:
                raise u.UnitConversionError(f'Column {col} has no units')
            self.ranges[col] <<= unit
        self._mask_auto_correlations = mask_auto_correlations

    @property
    def mask_auto_correlations(self) -> bool:
        return self._mask_auto_correlations

    def is_masked(self, frequency: u.Quantity, baseline_length: u.Quantity,
                  channel_width: u.Quantity = 0 * u.Hz) -> Any:
        # Add extra axis which will broadcast with the masks
        low = (frequency - 0.5 * channel_width)[..., np.newaxis]
        high = (frequency + 0.5 * channel_width)[..., np.newaxis]
        b = baseline_length[..., np.newaxis]
        in_range = (
            (self.ranges['min_frequency'] <= high)
            & (low <= self.ranges['max_frequency'])
            & (b <= self.ranges['max_baseline'])
        )
        if not self.mask_auto_correlations:
            in_range &= b > 0
        return np.any(in_range, axis=-1)

    def max_baseline_length(self, frequency: u.Quantity,
                            channel_width: u.Quantity = 0 * u.Hz) -> Any:
        # Add extra axis which will broadcast with the masks, and compute channel edges
        low = (frequency - 0.5 * channel_width)[..., np.newaxis]
        high = (frequency + 0.5 * channel_width)[..., np.newaxis]
        in_range = (self.ranges['min_frequency'] <= high) & (low <= self.ranges['max_frequency'])
        b = np.broadcast_to(self.ranges['max_baseline'], in_range.shape, subok=True)
        return np.max(b,
                      axis=-1, where=in_range,
                      initial=-1.0 * self.ranges['max_baseline'].unit)

    @classmethod
    def from_hdf5(cls: Type[_R], hdf5: h5py.File) -> _R:
        dataset_name = 'ranges'
        data = models.get_hdf5_dataset(hdf5, dataset_name)
        expected_dtype = np.dtype([
            ('min_frequency', 'f8'),
            ('max_frequency', 'f8'),
            ('max_baseline', 'f8')
        ])
        data = models.require_columns(f'{dataset_name} dataset', data, expected_dtype, 1)
        ranges = astropy.table.Table(data[...], copy=False)
        ranges['min_frequency'].unit = u.Hz
        ranges['max_frequency'].unit = u.Hz
        ranges['max_baseline'].unit = u.m
        mask_auto_correlations = models.get_hdf5_attr(
            hdf5.attrs, 'mask_auto_correlations', bool, required=True)
        return cls(ranges, mask_auto_correlations)

    def to_hdf5(self, hdf5: h5py.File) -> None:
        hdf5.attrs['mask_auto_correlations'] = self.mask_auto_correlations
        # The constructor ensures we're using unscaled units
        hdf5.create_dataset('ranges', data=self.ranges.as_array(), track_times=False)



################################################################################
# Copyright (c) 2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpmodels.band_mask`"""

import io
from typing import Generator

import astropy.units as u
import numpy as np
import pytest

from katsdpmodels import models, band_mask
import katsdpmodels.fetch.requests as fetch_requests
from test_utils import get_data_url


@pytest.fixture
def spectral_window() -> band_mask.SpectralWindow:
    return band_mask.SpectralWindow(400 * u.MHz, 1000 * u.MHz)


@pytest.fixture
def ranges_model(web_server) -> Generator[band_mask.BandMask, None, None]:
    with fetch_requests.fetch_model(web_server('band_mask_ranges.h5'), band_mask.BandMask) as model:
        yield model


def test_spectral_window_attributes(spectral_window) -> None:
    assert spectral_window.bandwidth == 400 * u.MHz
    assert spectral_window.centre_frequency == 1000 * u.MHz
    assert spectral_window.min_frequency == 800 * u.MHz
    assert spectral_window.max_frequency == 1200 * u.MHz


@pytest.mark.parametrize(
    'frequency,result',
    [
        (801 * u.MHz, True),
        (810 * u.MHz, True),
        (819 * u.MHz, True),
        (821 * u.MHz, False),
        (1000 * u.MHz, False),
        (1179 * u.MHz, False),
        (1181 * u.MHz, True),
        (1199 * u.MHz, True),
        (1201 * u.MHz, False),
        (800e6 * u.Hz, True),
        (1000e6 * u.Hz, False),
        (1200e6 * u.Hz, True)
    ])
def test_is_masked_scalar(frequency: u.Quantity, result: bool,
                          ranges_model: band_mask.BandMask,
                          spectral_window: band_mask.SpectralWindow) -> None:
    # Window is 800-1200 MHz, mask is 800-820 and 1180-1200
    assert ranges_model.is_masked(spectral_window, frequency) == result


def test_is_masked_vector(ranges_model: band_mask.BandMask,
                          spectral_window: band_mask.SpectralWindow) -> None:
    frequency = [801, 810, 819, 821, 1000, 1179, 1181, 1199, 1201] * u.MHz
    np.testing.assert_array_equal(
        ranges_model.is_masked(spectral_window, frequency),
        [True, True, True, False, False, False, True, True, False]
    )


def test_is_masked_channel_width(ranges_model: band_mask.BandMask,
                                 spectral_window: band_mask.SpectralWindow) -> None:
    frequency = [817, 819, 821, 823, 1177, 1179, 1181] * u.MHz
    np.testing.assert_array_equal(
        ranges_model.is_masked(spectral_window, frequency, 2 * u.MHz),
        [True, True, True, False, False, True, True]
    )


def test_bad_shape(mock_responses) -> None:
    url = get_data_url('band_mask_ranges_2d.h5')
    with pytest.raises(models.DataError,
                       match='ranges dataset should be 1-dimensional, but is 2-dimensional'):
        fetch_requests.fetch_model(url, band_mask.BandMask)


def test_bad_model_format(mock_responses) -> None:
    url = get_data_url('band_mask_bad_format.h5')
    with pytest.raises(models.ModelFormatError) as exc_info:
        fetch_requests.fetch_model(url, band_mask.BandMask)
    assert str(exc_info.value) == "Unknown model_format 'not_ranges' for band_mask"


def test_to_file(ranges_model: band_mask.BandMaskRanges) -> None:
    fh = io.BytesIO()
    ranges_model.to_file(fh, content_type='application/x-hdf5')
    fh.seek(0)
    new_model = band_mask.BandMask.from_file(fh, 'http://test.invalid/test.h5',
                                             content_type='application/x-hdf5')
    assert isinstance(new_model, band_mask.BandMaskRanges)
    np.testing.assert_array_equal(new_model.ranges, ranges_model.ranges)



################################################################################
# Copyright (c) 2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import asyncio
import concurrent.futures
import os
import pathlib
import threading
import urllib.parse
from typing import Tuple, Callable, Generator, Set

import pytest
import aioresponses
import responses
import katsdptelstate
import tornado.httpserver
import tornado.netutil
import tornado.web

import test_utils


_Info = Tuple[asyncio.AbstractEventLoop, asyncio.Event, str]


def _test_files() -> Generator[Tuple[str, str, bytes], None, None]:
    data_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')
    for (dirpath, dirnames, filenames) in os.walk(data_dir):
        for name in filenames:
            path = pathlib.Path(dirpath) / name
            rel_url = path.relative_to(data_dir).as_posix()
            url = urllib.parse.urljoin(test_utils.BASE_URL, rel_url)
            with open(path, 'rb') as f:
                data = f.read()
            content_type = 'application/octet-stream'
            if name.endswith('.h5'):
                content_type = 'application/x-hdf5'
            elif name.endswith('.alias'):
                content_type = 'text/plain'
            yield url, content_type, data


@pytest.fixture
def mock_responses() -> Generator[responses.RequestsMock, None, None]:
    """Fixture to make test data available via mocked HTTP (for :mod:`requests`)."""
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        for url, content_type, data in _test_files():
            rsps.add(responses.GET, url, content_type=content_type, body=data)
            rsps.add(responses.HEAD, url, content_type=content_type)
        yield rsps


@pytest.fixture
def mock_aioresponses() -> Generator[aioresponses.aioresponses, None, None]:
    """Fixture to make test data available via mocked HTTP (for :mod:`aiohttp`)."""
    with aioresponses.aioresponses() as rsps:
        for url, content_type, data in _test_files():
            rsps.get(url, content_type=content_type, body=data)
        yield rsps


class _FailOnceHandler(tornado.web.RequestHandler):
    def initialize(self, failed: Set[str]) -> None:
        self.failed = failed

    def get(self, path: str) -> None:
        if path not in self.failed:
            self.failed.add(path)
            raise tornado.web.HTTPError(500, 'Test error handling')
        else:
            self.redirect(f'/static/{path}')


@pytest.fixture
def web_server() -> Generator[Callable[[str], str], None, None]:
    """Fixture that runs a Tornado web server in a separate thread.

    It makes the test data available. The return value is a function that
    converts a relative URL to an absolute one for the server.
    """
    async def server_coro(info_future: 'concurrent.futures.Future[_Info]') -> None:
        try:
            data_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')
            app = tornado.web.Application(
                [
                    (r"/failonce/(.*)", _FailOnceHandler, dict(failed=set()))
                ],
                static_path=data_dir
            )
            sockets = tornado.netutil.bind_sockets(0, '127.0.0.1')
            port = sockets[0].getsockname()[1]
            server = tornado.httpserver.HTTPServer(app)
            server.add_sockets(sockets)
            server.start()
            finished_event = asyncio.Event()
            data_url = f'http://127.0.0.1:{port}/static/'
            info_future.set_result((asyncio.get_event_loop(), finished_event, data_url))
            await finished_event.wait()
            server.stop()
        except Exception as exc:
            info_future.set_exception(exc)

    def server_thread(info_future: 'concurrent.futures.Future[_Info]') -> None:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(server_coro(info_future))
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()

    def generate_url(relative: str) -> str:
        return urllib.parse.urljoin(base_url, relative)

    info_future: concurrent.futures.Future[_Info] = concurrent.futures.Future()
    thread = threading.Thread(target=server_thread, args=(info_future,))
    thread.start()
    # Wait for the server to be ready to serve
    loop, event, base_url = info_future.result()

    yield generate_url

    loop.call_soon_threadsafe(event.set)
    thread.join()


@pytest.fixture
def telstate() -> katsdptelstate.TelescopeState:
    """Telescope state with some attributes populated."""
    telstate = katsdptelstate.TelescopeState()
    telstate['sdp_model_base_url'] = test_utils.BASE_URL
    telstate['model_key'] = 'rfi_mask_ranges.h5'
    return telstate



################################################################################
# Copyright (c) 2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpmodels.models`."""

from datetime import datetime, timezone
import hashlib
import io
import pathlib
from typing import Dict, Type, Union

import numpy as np
import pytest

from katsdpmodels import models
from test_utils import DummyModel


@pytest.fixture
def dummy_model() -> DummyModel:
    ranges = np.array(
       [(1, 4.5), (2, -5.5)],
       dtype=[('a', 'i4'), ('b', 'f8')]
    )
    model = DummyModel(ranges)
    model.author = 'Test author'
    model.comment = 'Test comment'
    model.target = 'Test target'
    model.version = 1
    model.created = datetime(2020, 6, 15, 14, 11, tzinfo=timezone.utc)
    return model


def test_eq_hash() -> None:
    # model1 and model2 have the same checksum, model3 a different checksum
    model1 = DummyModel(None)
    model2 = DummyModel(None)
    model3 = DummyModel(None)
    model4 = DummyModel(None)
    model1.checksum = hashlib.sha256(b'foo').hexdigest()
    model2.checksum = model1.checksum
    model3.checksum = hashlib.sha256(b'bar').hexdigest()
    assert model1 == model1
    assert model1 == model2
    assert model1 != model3
    assert model3 == model3
    assert model3 != model4
    assert model4 == model4
    assert hash(model1) == hash(model2)
    assert hash(model1) != hash(model3)
    assert hash(model3) != hash(model4)
    assert model1 != 1
    assert model4 != 1


def test_get_hdf5_attr_missing() -> None:
    attrs: Dict[str, object] = {}
    assert models.get_hdf5_attr(attrs, 'missing', str) is None
    with pytest.raises(KeyError):
        models.get_hdf5_attr({}, 'missing', str, required=True)


def test_get_hdf5_attr_type_mismatch() -> None:
    with pytest.raises(TypeError, match="Expected <class 'int'> for 'foo', received <class 'str'>"):
        models.get_hdf5_attr({'foo': 'bar'}, 'foo', int)


def test_get_hdf5_attr_decode_bytes() -> None:
    assert models.get_hdf5_attr({'foo': 'café'.encode()}, 'foo', str) == 'café'


def test_get_hdf5_attr_numpy_int() -> None:
    assert models.get_hdf5_attr({'foo': np.int64(1)}, 'foo', int) == 1


def test_get_hdf5_attr_bad_utf8() -> None:
    with pytest.raises(UnicodeDecodeError):
        models.get_hdf5_attr({'foo': b'\xff'}, 'foo', str)
    # Should raise the type error first if we aren't expecting strings
    with pytest.raises(TypeError):
        models.get_hdf5_attr({'foo': b'\xff'}, 'foo', int)


def test_get_hdf5_attr_bool_not_int() -> None:
    with pytest.raises(TypeError):
        models.get_hdf5_attr({'foo': True}, 'foo', int)
    with pytest.raises(TypeError):
        models.get_hdf5_attr({'foo': 2}, 'foo', bool)


def test_get_hdf5_attr_success() -> None:
    attrs = {'string': 'hello', 'int': 42}
    assert models.get_hdf5_attr(attrs, 'string', str) == 'hello'
    assert models.get_hdf5_attr(attrs, 'int', int) == 42


def test_require_columns_missing_column() -> None:
    dtype1 = np.dtype([('a', 'f8'), ('b', 'i4')])
    dtype2 = np.dtype([('a', 'f8'), ('c', 'i4')])
    array = np.zeros((5,), dtype1)
    with pytest.raises(models.DataError, match='foo column c is missing'):
        models.require_columns('foo', array, dtype2, 1)


def test_require_columns_unstructured() -> None:
    dtype = np.dtype([('a', 'f8'), ('b', 'i4')])
    array = np.zeros((5,), np.float32)
    with pytest.raises(models.DataError, match='foo does not have named columns'):
        models.require_columns('foo', array, dtype, 1)


def test_require_columns_dtype_mismatch() -> None:
    dtype1 = np.dtype([('a', 'f8')])
    dtype2 = np.dtype([('a', 'i4')])
    array = np.zeros((5,), dtype1)
    with pytest.raises(models.DataError, match='foo column a has type float64, expected int32'):
        models.require_columns('foo', array, dtype2, 1)


def test_require_columns_same_dtype() -> None:
    dtype = np.dtype([('a', 'f8'), ('b', 'i4')])
    array = np.array([(1.5, 1), (3.5, 3)], dtype=dtype)
    out = models.require_columns('foo', array, dtype, 1)
    np.testing.assert_array_equal(array, out)
    assert np.shares_memory(array, out)


def test_require_columns_change_byteorder() -> None:
    dtype1 = np.dtype([('a', '>f8')])
    dtype2 = np.dtype([('a', '<f8')])
    array = np.array([1.0, 1.5, 2.0], dtype=dtype1)
    out = models.require_columns('foo', array, dtype2, 1)
    np.testing.assert_array_equal(array, out)
    # Some versions of numpy considered dtypes to be equal even if the byte
    # order was different, so explicitly compare byte order.
    assert out.dtype['a'].byteorder == dtype2['a'].byteorder


def test_require_columns_extra_column() -> None:
    dtype1 = np.dtype([('a', 'f8'), ('b', 'f8')])
    dtype2 = np.dtype([('b', 'f8')])
    array = np.array([(1.0, 10.0), (1.5, 15.0), (2.0, 20.0)], dtype=dtype1)
    expected = np.array([10.0, 15.0, 20.0], dtype=dtype2)
    out = models.require_columns('foo', array, dtype2, 1)
    np.testing.assert_equal(out, expected)


def test_require_columns_dimension_mismatch() -> None:
    dtype = np.dtype([('a', 'f8'), ('b', 'i4')])
    array = np.zeros((1, 2), dtype)
    with pytest.raises(models.DataError,
                       match='foo should be 1-dimensional, but is 2-dimensional'):
        models.require_columns('foo', array, dtype, 1)


def assert_models_equal(model1: DummyModel, model2: DummyModel):
    assert type(model1) == type(model2)
    assert model1.model_type == model2.model_type
    assert model1.model_format == model2.model_format
    assert model1.comment == model2.comment
    assert model1.author == model2.author
    assert model1.target == model2.target
    assert model1.created == model2.created
    assert model1.version == model2.version
    np.testing.assert_array_equal(model1.ranges, model2.ranges)


@pytest.mark.parametrize('clear_metadata', [False, True])
def test_hdf5_to_file(clear_metadata: bool, dummy_model: DummyModel) -> None:
    if clear_metadata:
        dummy_model.comment = None
        dummy_model.author = None
        dummy_model.target = None
        dummy_model.created = None
    fh = io.BytesIO()
    dummy_model.to_file(fh, content_type='application/x-hdf5')
    fh.seek(0)
    new_model = DummyModel.from_file(fh, 'http://test.invalid/dummy.h5',
                                     content_type='application/x-hdf5')
    assert_models_equal(dummy_model, new_model)


def test_hdf5_to_file_no_version(dummy_model: DummyModel) -> None:
    dummy_model.version = None
    with pytest.raises(ValueError):
        dummy_model.to_file(io.BytesIO(), content_type='application/x-hdf5')


def test_hdf5_to_file_no_content_type_or_filename(dummy_model: DummyModel) -> None:
    with pytest.raises(AttributeError):
        dummy_model.to_file(io.BytesIO())


def test_hdf5_to_file_bad_content_type(dummy_model: DummyModel) -> None:
    with pytest.raises(models.FileTypeError, match='Expected application/x-hdf5, not image/png'):
        dummy_model.to_file(io.BytesIO(), content_type='image/png')


@pytest.mark.parametrize('path_type', [pathlib.Path, str])
def test_hdf5_to_file_bad_extension(path_type: Union[Type[pathlib.Path], Type[str]],
                                    dummy_model: DummyModel,
                                    tmp_path: pathlib.Path) -> None:
    with pytest.raises(models.FileTypeError,
                       match=r'Expected extension of \.h5 or \.hdf5, not \.foo'):
        dummy_model.to_file(path_type(tmp_path / 'test.foo'))


@pytest.mark.parametrize('path_type', [pathlib.Path, str])
def test_hdf5_to_file_good_extension(path_type: Union[Type[pathlib.Path], Type[str]],
                                     dummy_model: DummyModel,
                                     tmp_path: pathlib.Path) -> None:
    path = path_type(tmp_path / 'test.h5')
    dummy_model.to_file(path)
    new_model = dummy_model.from_file(open(path, 'rb'), 'test.h5',
                                      content_type='application/x-hdf5')
    assert_models_equal(dummy_model, new_model)
